from pathlib import Path
from typing import Callable, Union

import numpy as np
from spinesUtils.asserts import raise_if
from spinesUtils.logging import Logger

from ..core_components.locks import ThreadLock
from ..storage_layer.storage import PersistentFileStorage
from ..index.sq import IndexSQIP, IndexSQL2sq, IndexSQCos
from ..index.binary import IndexBinaryJaccard, IndexBinaryHamming
from ..index.flat import IndexFlatIP, IndexFlatL2sq, IndexFlatCos
from .ivf import IVFCreator
from ..utils.utils import drop_duplicated_substr, find_first_file_with_substr

_INDEX_ALIAS = {
    'IVF-IP-SQ8': 'IVF-IP-SQ8',
    'IVF': 'IVF-IP',
    'IVF-L2sq-SQ8': 'IVF-L2sq-SQ8',
    'IVF-L2sq': 'IVF-L2sq',
    'IVF-Cos-SQ8': 'IVF-Cos-SQ8',
    'IVF-Cos': 'IVF-Cos',
    'IVF-Jaccard-Binary': 'IVF-Jaccard-Binary',
    'IVF-Hamming-Binary': 'IVF-Hamming-Binary',
    'Flat-IP-SQ8': 'Flat-IP-SQ8',
    'FLAT': 'Flat-IP',
    'Flat-L2sq-SQ8': 'Flat-L2sq-SQ8',
    'Flat-L2sq': 'Flat-L2sq',
    'Flat-Cos-SQ8': 'Flat-Cos-SQ8',
    'Flat-Cos': 'Flat-Cos',
    'Flat-Hamming-Binary': 'Flat-Hamming-Binary',
    'Flat-Jaccard-Binary': 'Flat-Jaccard-Binary'
}

_IndexMapper = {
    'IVF-IP-SQ8': (IndexSQIP, IVFCreator),
    'IVF-IP': (IndexFlatIP, IVFCreator),
    'IVF-L2sq-SQ8': (IndexSQL2sq, IVFCreator),
    'IVF-L2sq': (IndexFlatL2sq, IVFCreator),
    'IVF-Cos-SQ8': (IndexSQCos, IVFCreator),
    'IVF-Cos': (IndexFlatCos, IVFCreator),
    'IVF-Jaccard-Binary': (IndexBinaryJaccard, IVFCreator),
    'IVF-Hamming-Binary': (IndexBinaryHamming, IVFCreator),
    'Flat-IP-SQ8': (IndexSQIP, None),
    'Flat-IP': (IndexFlatIP, None),
    'Flat-L2sq-SQ8': (IndexSQL2sq, None),
    'Flat-L2sq': (IndexFlatL2sq, None),
    'Flat-Cos-SQ8': (IndexSQCos, None),
    'Flat-Cos': (IndexFlatCos, None),
    'Flat-Jaccard-Binary': (IndexBinaryJaccard, None),
    'Flat-Hamming-Binary': (IndexBinaryHamming, None)
}


class Indexer:
    def __init__(
            self,
            logger: Logger,
            dataloader: Callable,
            storage_worker: PersistentFileStorage,
            collections_path_parent: Union[str, Path],
    ):
        self.logger = logger
        self.dataloader = dataloader
        self.ann_model = None
        self.storage_worker = storage_worker
        self.collections_path_parent = Path(collections_path_parent)
        self.index_data_path = self.collections_path_parent / 'index_data'
        self.index_ids_path = self.collections_path_parent / 'index_ids'
        self.index_path = self.collections_path_parent / 'index'

        for _ in [self.index_data_path, self.index_ids_path, self.index_path]:
            _.mkdir(parents=True, exist_ok=True)

        self.lock = ThreadLock()

        self.index = None
        self.ivf = None
        self.index_mode = None
        self.current_index_mode = None

    def _build_flat_index(self, index_mode: str):
        """
        Build the flat index.

        Parameters:
            index_mode (str): The index mode.
        """
        def _rebuild():
            nonlocal _index

            filenames = self.storage_worker.get_all_files()
            for filename in filenames:
                data, ids = self.dataloader(filename)
                _index.fit_transform(data, ids)

            _index.save(self.index_path / f'{index_mode}.index')
            if hasattr(_index, "save_data"):
                if index_mode.endswith('SQ8'):
                    data_path = self.index_data_path / f'{self.storage_worker.fingerprint}.sqd'
                    ids_path = self.index_ids_path / f'{self.storage_worker.fingerprint}.sqi'
                elif 'Binary' in index_mode:
                    data_path = self.index_data_path / f'{self.storage_worker.fingerprint}.bd'
                    ids_path = self.index_ids_path / f'{self.storage_worker.fingerprint}.bi'
                else:
                    raise NotImplementedError

                _index.save_data(data_path, ids_path)
            return _index

        def build_sq8_index_for_binary():
            index = _IndexMapper['Flat-IP-SQ8'][0]()
            fitted = False
            for index_name in self.index_path.iterdir():
                if 'SQ8' in index_name.name:
                    index = index.load(index_name)
                    fitted = True
                    break

            if not fitted:
                filenames = self.storage_worker.get_all_files()
                _data = []
                _ids = []
                for filename in filenames:
                    data, ids = self.dataloader(filename)
                    data = np.vstack([np.vstack(_data), data]) if _data else data
                    ids = np.hstack([np.hstack(_ids), ids]) if _ids else ids
                    if data.shape[0] < 100000:
                        _data.append(data)
                        _ids.append(ids)
                        continue

                    _data, _ids = [], []
                    index.fit_transform(data, ids)
                index.save(self.index_path / 'Flat-IP-SQ8.index')
                index.save_data(self.index_data_path / f'{self.storage_worker.fingerprint}.sqd',
                                self.index_ids_path / f'{self.storage_worker.fingerprint}.sqi')
                # clear the data and ids, because they are saved in the file
                index.data = None
                index.ids = None

            return index

        def build_binary():
            nonlocal _index
            if not (self.index_data_path / f'{self.storage_worker.fingerprint}.bd').exists():
                _index = _rebuild()
            else:
                with open(self.index_data_path / f'{self.storage_worker.fingerprint}.bd', 'rb') as file:
                    binary_data = np.load(file)
                with open(self.index_ids_path / f'{self.storage_worker.fingerprint}.bi', 'rb') as file:
                    binary_ids = np.load(file)

                # load sq8 data as a view, used for rescore
                if (not (self.index_data_path / f'{self.storage_worker.fingerprint}.sqd').exists()) or (
                        find_first_file_with_substr(
                            self.index_ids_path,
                            f'{self.storage_worker.fingerprint}.*SQ8.index') is None
                ):
                    sq8_index = build_sq8_index_for_binary()
                else:
                    sq8_index = _IndexMapper['Flat-IP-SQ8'][0]()

                    sq8_index = sq8_index.load(
                        find_first_file_with_substr(self.index_ids_path,
                                                    f'{self.storage_worker.fingerprint}.*SQ8.index')
                    )

                sq8_data = np.load(self.index_data_path / f'{self.storage_worker.fingerprint}.sqd',
                                   mmap_mode='r')

                _index.data = binary_data
                _index.ids = binary_ids
                _index.sq8_data = sq8_data
                _index.sq8_encode = sq8_index.encode
                _index.sq8_decode = sq8_index.decode

            return _index

        # ========== main logic ==========
        uninitialized_index, _ = _IndexMapper[index_mode]
        _index = uninitialized_index()

        if (self.index_path / f'{index_mode}.index').exists():
            _index = _index.load(self.index_path / f'{index_mode}.index')
        else:
            _index = _rebuild()

        if 'Binary' in index_mode:
            _index = build_binary()
        elif 'SQ8' in index_mode:
            if (not (self.index_data_path / f'{self.storage_worker.fingerprint}.sqd').exists()) or (
                not (self.index_ids_path / f'{self.storage_worker.fingerprint}.sqi').exists()
            ):
                _index = _rebuild()
            else:
                with open(self.index_data_path / f'{self.storage_worker.fingerprint}.sqd', 'rb') as file:
                    sq8_data = np.load(file)
                with open(self.index_ids_path / f'{self.storage_worker.fingerprint}.sqi', 'rb') as file:
                    ids = np.load(file)

                _index.data = sq8_data
                _index.ids = ids
        else:
            ...  # do nothing

        self.index = _index
        self.index_mode = index_mode

    def _build_ivf_index(self, n_clusters: int, rebuild=False):
        """
        Build the IVF index.

        Parameters:
            n_clusters (int): The number of clusters.
            rebuild (bool): Whether to rebuild the index.
        """
        self.ivf = IVFCreator(
            dataloader=self.dataloader,
            storage_worker=self.storage_worker,
            collections_path_parent=self.collections_path_parent
        )

        self.ivf.build_index(n_clusters=n_clusters, rebuild=rebuild)

    def build_index(self, index_mode='IVF-IP-SQ8', rebuild=False, **kwargs):
        """
        Build the index for clustering.

        Parameters:
            index_mode (str): The index mode, must be one of the following:

                - 'IVF-IP-SQ8': IVF index with inner product and scalar quantizer with 8 bits.
                    The distance is inner product.
                - 'IVF-IP': IVF index with inner product. (Alias: 'IVF')
                - 'IVF-L2sq-SQ8': IVF index with squared L2 distance and scalar quantizer with 8 bits.
                    The distance is squared L2 distance.
                - 'IVF-L2sq': IVF index with squared L2 distance.
                - 'IVF-Cos-SQ8': IVF index with cosine similarity and scalar quantizer with 8 bits.
                    The distance is cosine similarity.
                - 'IVF-Cos': IVF index with cosine similarity.
                - 'IVF-Jaccard-Binary': IVF index with binary quantizer. The distance is Jaccard distance.
                - 'IVF-Hamming-Binary': IVF index with binary quantizer. The distance is Hamming distance.
                - 'Flat-IP-SQ8': Flat index with inner product and scalar quantizer with 8 bits.
                - 'Flat-IP': Flat index with inner product. (Alias: 'FLAT')
                - 'Flat-L2sq-SQ8': Flat index with squared L2 distance and scalar quantizer with 8 bits.
                - 'Flat-L2sq': Flat index with squared L2 distance.
                - 'Flat-Cos-SQ8': Flat index with cosine similarity and scalar quantizer with 8 bits.
                - 'Flat-Cos': Flat index with cosine similarity.
                - 'Flat-Jaccard-Binary': Flat index with binary quantizer. The distance is Jaccard distance.
                - 'Flat-Hamming-Binary': Flat index with binary quantizer. The distance is Hamming distance.
            rebuild (bool): Whether to rebuild the index.
            kwargs: Additional keyword arguments. The following are available:

                    - 'n_clusters': The number of clusters. It is only available when the index_mode including 'IVF'.

        Returns:
            None
        """
        # clear the old data first
        self._remove_old_data()

        n_clusters = kwargs.setdefault('n_clusters', 32)
        raise_if(ValueError, not isinstance(n_clusters, int) or n_clusters <= 0,
                 'n_clusters must be int and greater than 0')

        raise_if(ValueError, index_mode not in _INDEX_ALIAS and index_mode not in _INDEX_ALIAS.values(),
                 'index_mode must be one of the following:'
                 f' {list(_INDEX_ALIAS) + list(_INDEX_ALIAS.values())}')

        if (not rebuild) and self.index_mode == index_mode:
            self.logger.info('Index already exists.')
            return

        if index_mode in _INDEX_ALIAS:
            index_mode = _INDEX_ALIAS[index_mode]

        self.logger.info(f'Building an index using the `{index_mode}` index mode...')
        all_partition_size = self.storage_worker.get_shape()[0]

        if 'Binary' in index_mode or 'SQ8' in index_mode:
            if all_partition_size < 100000:
                substr = '-Binary' if 'Binary' in index_mode else '-SQ8'
                index_mode = drop_duplicated_substr(index_mode, substr).replace("IVF", "Flat")
                self.logger.info('Index is not built because the number of data points is less than 100000.'
                                 f'Continue to build the {index_mode} index.')

        with self.lock:
            self._build_flat_index(index_mode)
        if not index_mode.startswith('Flat'):
            with self.lock:
                self._build_ivf_index(n_clusters, rebuild=rebuild)

        self.current_index_mode = index_mode
        self.logger.info('Index built.')

    def index_insert(self, data, ids):
        """
        Insert the data into the index.

        Parameters:
            data (np.ndarray): The data to insert.
            ids (np.ndarray): The IDs of the data.

        Returns:
            None
        """
        raise_if(ValueError, self.index is None, 'The index must be built before inserting data.')

        with self.lock:
            self.index.fit_transform(data, ids)

    def update_filenames(self):
        # get new fingerprint
        with open(self.storage_worker.fingerprint_path, 'r') as file:
            new_fingerprint = file.readlines()[-1].strip()

        self._remove_old_data()
        if hasattr(self.index, 'save_data'):
            self.index.save_data(self.index_data_path / f'{new_fingerprint}.sqd',
                                 self.index_ids_path / f'{new_fingerprint}.sqi')

    def ivf_insert(self, last_filename):
        self.ivf.fast_insert(last_filename)

    def _remove_old_data(self):
        """
        Remove old data, including the index and the data.
        """
        # read all history fingerprints
        if not self.storage_worker.fingerprint_path.exists():
            return
        with open(self.storage_worker.fingerprint_path, 'r') as file:
            old_fingerprints = file.readlines()[:-1]

        if old_fingerprints:
            for fingerprint in old_fingerprints:
                fingerprint = fingerprint.strip()
                for _ in [self.index_data_path, self.index_ids_path, self.index_path]:
                    for file in _.iterdir():
                        if fingerprint in file.name:
                            file.unlink()

    def remove_index(self):
        """
        Remove the index.

        Returns:
            None
        """
        self.index = None
        self.index_mode = None
        # remove local index
        if self.current_index_mode is not None:
            if self.current_index_mode.startswith('IVF'):
                self.ivf.remove_index()
                self.ivf = None
            else:
                index_mode = self.current_index_mode
                if index_mode in _INDEX_ALIAS:
                    index_mode = _INDEX_ALIAS[index_mode]
                if (self.index_path / f'{index_mode}_index').exists():
                    (self.index_path / f'{index_mode}_index').unlink()

        self.current_index_mode = 'Flat-IP'  # Default index mode

        self.logger.info('Index removed.')



import threading


class ThreadLock:
    def __init__(self):
        self._thread_lock = threading.RLock()

    def acquire(self):
        """Acquire the lock based on the running environment."""
        self._thread_lock.acquire()

    def release(self):
        """Release the lock based on the running environment."""
        self._thread_lock.release()

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()



import os.path
from pathlib import Path

import numpy as np

import lynse


client = lynse.VectorDBClient()


def test_create_collection():
    my_vec_db = client.create_database("my_vec_db", drop_if_exists=False)
    collection = my_vec_db.require_collection('test_collection', dim=4, drop_if_exists=True)

    assert Path(collection._database_path).stem == 'test_collection'
    assert collection._matrix_serializer.dim == 4
    assert collection._matrix_serializer.chunk_size == 100000
    assert collection._matrix_serializer.dtypes == np.float32
    assert Path(os.path.expanduser("~/.LynseDB/databases/my_vec_db/test_collection")).exists()

    my_vec_db.drop_collection('test_collection')
    assert not Path(os.path.expanduser("~/.LynseDB/databases/my_vec_db/test_collection")).exists()

    my_vec_db.drop_database()
    assert not Path(os.path.expanduser("~/.LynseDB/databases/my_vec_db/test_collection/my_vec_db")).exists()


def test_show_collections():
    my_vec_db = client.create_database("my_vec_db", drop_if_exists=False)
    collection = my_vec_db.require_collection('test_collection', dim=4, drop_if_exists=True)

    collections = my_vec_db.show_collections()
    assert collections == ['test_collection']

    my_vec_db.drop_collection('test_collection')
    my_vec_db.drop_database()
    assert not Path(os.path.expanduser("~/.LynseDB/databases/my_vec_db/test_collection/my_vec_db")).exists()


def test_get_an_exists_collection():
    my_vec_db = client.create_database("my_vec_db", drop_if_exists=False)
    collection = my_vec_db.require_collection('test_collection', dim=4, drop_if_exists=True)

    collection = my_vec_db.get_collection('test_collection')
    assert Path(collection._database_path).stem == 'test_collection'

    my_vec_db.drop_collection('test_collection')
    my_vec_db.drop_database()
    assert not Path(os.path.expanduser("~/.LynseDB/databases/my_vec_db/test_collection/my_vec_db")).exists()



import operator
from typing import List

from . import ExpressionParser
from ...core_components.fields_cache.filter import Filter, MatchField, FieldCondition


class FieldsQuery:
    def __init__(self, storage):
        self.storage = storage

    def filter_non_id_conditions(self, condition_list):
        return [condition for condition in condition_list
                if condition.matcher.name != "MatchID" and condition.key != ":id:"]

    def filter_range_conditions(self, condition_list):
        return [condition for condition in condition_list
                if condition.matcher.name == "MatchRange" or
                (condition.matcher.name == "MatchField" and condition.matcher.comparator in [operator.le, operator.lt,
                                                                                             operator.ge, operator.gt])]

    def filter_equal_conditions(self, condition_list):
        return [condition for condition in condition_list
                if
                condition.matcher.name == "MatchField" and condition.matcher.comparator in [operator.eq, operator.ne]]

    def retrieve_all_ids(self):
        return self.storage.retrieve_all_ids()

    def is_filter_empty(self, filter_instance):
        return not filter_instance.must_fields and \
            not filter_instance.must_not_fields and \
            not filter_instance.any_fields

    def _process_conditions(self, range_conditions=None, equal_conditions=None, is_any_fields=False):
        """
        Process the conditions.
        """
        if not range_conditions and not equal_conditions:
            return set()

        match_ids = set()
        non_indexed_conditions = []

        # handle range conditions
        for condition in range_conditions:
            if condition.matcher.name == 'MatchRange':
                start = condition.matcher.start
                end = condition.matcher.end
                inclusive = condition.matcher.inclusive

                end = end if inclusive != "left" and inclusive is not False else end - 1
                start = start if inclusive != "right" and inclusive is not False else start + 1

                if condition.key in self.storage.index.indices:
                    match_ids |= self.storage.index.range_search(condition.key, start, end)
                else:
                    non_indexed_conditions.append(condition)
            elif condition.matcher.name == 'MatchField':
                comparator = getattr(condition.matcher, 'comparator', None)
                if comparator is None:
                    raise ValueError("The comparator must be provided for MatchField conditions")

                start_value = None
                end_value = None
                if comparator == operator.lt:
                    end_value = condition.matcher.value - 1
                elif comparator == operator.le:
                    end_value = condition.matcher.value
                elif comparator == operator.gt:
                    start_value = condition.matcher.value + 1
                else:
                    start_value = condition.matcher.value

                if condition.key in self.storage.index.indices:
                    match_ids |= self.storage.index.range_search(condition.key, start_value, end_value)
                else:
                    non_indexed_conditions.append(condition)

        if non_indexed_conditions:
            match_ids |= self._match_filter(non_indexed_conditions, is_any_fields)

        non_indexed_conditions = []
        # handle equal conditions
        for condition in equal_conditions:
            if condition.matcher.name == 'MatchField':
                comparator = getattr(condition.matcher, 'comparator', None)
                if comparator is None:
                    raise ValueError("The comparator must be provided for MatchField conditions")

                if condition.key in self.storage.index.indices:
                    match_ids |= self.storage.index.search(condition.key, condition.matcher.value)
                else:
                    non_indexed_conditions.append(condition)

        if non_indexed_conditions:
            match_ids |= self._match_filter(non_indexed_conditions, is_any_fields)

        return match_ids

    def _match_filter(self, conditions, is_any_fields=False):
        """
        Match the filter.
        """
        match_ids = set()
        compare_op = all if not is_any_fields else any
        for external_id, data in self.storage.retrieve_all():
            if compare_op(condition.evaluate(data, external_id) for condition in conditions):
                match_ids.add(external_id)

        return match_ids

    def _pre_match_ids(self, condition_list, is_any_fields=False):
        """
        Pre-match the IDs.
        """
        pre_match_ids = set()
        for condition in condition_list:
            if condition.matcher.name == "MatchID":
                if pre_match_ids:
                    pre_match_ids = pre_match_ids & set(condition.matcher.indices) \
                        if not is_any_fields else pre_match_ids | set(condition.matcher.indices)
                else:
                    pre_match_ids = set(condition.matcher.indices)
            elif condition.key == ":id:":
                if pre_match_ids:
                    pre_match_ids = pre_match_ids & set(self.storage.id_filter.filter_ids(condition)) \
                        if not is_any_fields else pre_match_ids | set(self.storage.id_filter.filter_ids(condition))
                else:
                    pre_match_ids = set(self.storage.id_filter.filter_ids(condition))
        return pre_match_ids

    def _query(self, filter_instance, filter_ids=None, return_ids_only=True):
        """
        Query the fields cache using normal method.
        """
        if self.is_filter_empty(filter_instance):
            match_ids = self.storage.id_filter.retrieve_all_ids(return_set=True)
        else:
            match_ids = set()
            not_match_ids = set()
            any_match_ids = set()

            must_fields = filter_instance.must_fields
            must_not_fields = filter_instance.must_not_fields
            any_fields = filter_instance.any_fields

            # if has must fields, filter by match id
            if must_fields:
                match_ids.update(self._pre_match_ids(must_fields))
            else:
                match_ids = self.storage.id_filter.retrieve_all_ids(return_set=True)

            # if no must fields, filter by must_not fields
            if must_not_fields:
                not_match_ids.update(self._pre_match_ids(must_not_fields))

            if any_fields:
                any_match_ids.update(self._pre_match_ids(any_fields, is_any_fields=True))

            must_fields = self.filter_non_id_conditions(must_fields)
            must_not_fields = self.filter_non_id_conditions(must_not_fields)
            any_fields = self.filter_non_id_conditions(any_fields)

            if must_fields:
                # process the remaining conditions
                match_ids |= self._process_conditions(
                    self.filter_range_conditions(must_fields),
                    self.filter_equal_conditions(must_fields)
                )
            if must_not_fields:
                not_match_ids |= self._process_conditions(
                    self.filter_range_conditions(must_not_fields),
                    self.filter_equal_conditions(must_not_fields)
                )
            if any_fields:
                any_match_ids |= self._process_conditions(
                    self.filter_range_conditions(any_fields),
                    self.filter_equal_conditions(any_fields),
                    is_any_fields=True
                )

            if not_match_ids:
                match_ids -= not_match_ids

            if any_match_ids:
                # match_ids is never empty
                match_ids &= any_match_ids

            if not match_ids:
                return []

        if return_ids_only:
            return list(match_ids) if not filter_ids else list(match_ids - set(filter_ids))
        else:
            return self.retrieve_ids(list(match_ids), include_external_id=True)

    def query(self, query_filter, filter_ids=None, return_ids_only=True):
        """
        Query the fields cache.

        Parameters:
            query_filter: Filter or dict or FieldExpression (string)
                The filter object or the specify data to filter.
            filter_ids: List[int]
                The list of external IDs to filter.
            return_ids_only: bool
                If True, only the external IDs will be returned.

        Returns:
            List[dict]: The records. If not return_ids_only, the records will be returned.
            List[int]: The external IDs. If return_ids_only, the external IDs will be returned.
        """
        if isinstance(query_filter, dict):
            if ('must_fields' not in query_filter or
                    'must_not_fields' not in query_filter or
                    'any_fields' not in query_filter):
                final_filter = []
                for key, value in query_filter.items():
                    final_filter.append(FieldCondition(key=key, matcher=MatchField(value, comparator=operator.eq)))
                    query_filter = Filter(must=final_filter)
            else:
                query_filter = Filter().load_dict(query_filter)
        elif isinstance(query_filter, str):
            query_filter = ExpressionParser(query_filter).to_filter()
        elif not isinstance(query_filter, Filter):
            raise ValueError("Invalid query filter")

        return self._query(query_filter, filter_ids, return_ids_only)

    def retrieve(self, external_id, include_external_id=True):
        """
        Retrieve a record from the cache.

        Parameters:
            external_id: int
                The external ID of the record.
            include_external_id: bool
                If True, the external ID will be included in the record.

        Returns:
            dict: The record.
        """
        res = self.storage.retrieve_by_external_id(external_id, include_external_id=include_external_id)
        if res:
            return res[0]
        return None

    def retrieve_ids(self, external_ids: List[int], include_external_id=True):
        """
        Retrieve records from the cache.

        Parameters:
            external_ids: List[int]
                The external IDs of the records.
            include_external_id: bool
                If True, the external ID will be included in the record.

        Returns:
            List[dict]: The records.
        """
        results = self.storage.retrieve_by_external_id(external_ids, include_external_id=include_external_id)
        return results if results else None



import os
from spinesUtils.logging import Logger

from ..configs.config import config

# Create a logger for the LynseDB
logger = Logger(
    fp=os.environ.get('LYNSE_LOG_PATH') or config.LYNSE_LOG_PATH,
    name='LynseDB',
    truncate_file=os.environ.get('LYNSE_TRUNCATE_LOG_FILE') or config.LYNSE_TRUNCATE_LOG,
    with_time=os.environ.get('LYNSE_LOG_WITH_TIME') or config.LYNSE_LOG_WITH_TIME,
    level=os.environ.get('LYNSE_LOG_LEVEL') or config.LYNSE_LOG_LEVEL
)



import numpy as np
import simsimd

from usearch.compiled import exact_search
from usearch.index import MetricKind

from ..core_components.fast_sort import FastSort


def inner_product(vec1, vec2, n, use_simd=False):
    """
    Calculate the inner product between a vector and each row of a 2D matrix.

    Parameters:
        vec1 (np.ndarray): The vector.
        vec2 (np.ndarray): The 2D matrix.
        n (int): The number of vectors to return.

    Returns:
        (np.ndarray, np.ndarray): Indices of the top vectors and the result vector to store inner products.
    """
    vec1 = _check_first_input(vec1, vec2)
    if not use_simd and vec1.dtype != np.int8:
        vec1 = _check_first_input(vec1, vec2)
        ids, distance, *_ = exact_search(vec2, vec1, n, metric_kind=MetricKind.IP)
    else:
        dis = simsimd.cdist(vec1, vec2, "inner")
        topk = FastSort(dis)
        ids, distance = topk.topk(n, ascending=False)
    return _wrap_results(ids, distance)


def _check_first_input(vec1, vec2):
    if vec1.dtype != vec2.dtype:
        vec1 = vec1.astype(vec2.dtype)

    return np.atleast_2d(vec1)


def _wrap_results(ids, distance):
    return ids.squeeze(), distance.squeeze()


def l2sq(vec1, vec2, n, use_simd=True):
    vec1 = _check_first_input(vec1, vec2)
    if not use_simd:
        ids, distance, *_ = exact_search(vec2, vec1, n, metric_kind=MetricKind.L2sq)
    else:
        dis = simsimd.sqeuclidean(vec1, vec2)
        topk = FastSort(dis)
        ids, distance = topk.topk(n, ascending=True)

    return _wrap_results(ids, distance)


def cosine(vec1, vec2, n, use_simd=False):
    vec1 = _check_first_input(vec1, vec2)
    if not use_simd:
        ids, distance, *_ = exact_search(vec2, vec1, n, metric_kind=MetricKind.Cos)
    else:
        dis = simsimd.cdist(vec1, vec2, "cosine")
        topk = FastSort(dis)
        ids, distance = topk.topk(n, ascending=True)

    return _wrap_results(ids, distance)


def hamming(vec1, vec2, n, use_simd=False):
    vec1 = _check_first_input(vec1, vec2)
    if not use_simd:
        ids, distance, *_ = exact_search(vec2, vec1, n, metric_kind=MetricKind.Hamming)
    else:
        dis = simsimd.cdist(vec1, vec2, "hamming")
        topk = FastSort(dis)
        ids, distance = topk.topk(n, ascending=True)

    return _wrap_results(ids, distance)


def jaccard(vec1, vec2, n, use_simd=False):
    vec1 = _check_first_input(vec1, vec2)
    if not use_simd:
        ids, distance, *_ = exact_search(vec2, vec1, n, metric_kind=MetricKind.Jaccard)
    else:
        dis = simsimd.cdist(vec1, vec2, "jaccard")
        topk = FastSort(dis)
        ids, distance = topk.topk(n, ascending=True)

    return _wrap_results(ids, distance)


from pathlib import Path
from typing import Callable, Union

import numpy as np
from spinesUtils.asserts import raise_if

from ..configs.config import config
from ..storage_layer.storage import PersistentFileStorage
from ..core_components.kmeans import BatchKMeans
from ..core_components.ivf_index import IVFIndex
from ..core_components.locks import ThreadLock


class IVFCreator:
    def __init__(
            self,
            dataloader: Callable,
            storage_worker: PersistentFileStorage,
            collections_path_parent: Union[str, Path],
    ):
        self.dataloader = dataloader
        self.ann_model = None
        self.storage_worker = storage_worker
        self.ivf_index = IVFIndex()

        self.index_path = Path(collections_path_parent) / 'index'
        self.index_path.mkdir(parents=True, exist_ok=True)

        if (self.index_path / 'ivf_ann_model').exists():
            self.ann_model = BatchKMeans(n_clusters=1, batch_size=10240, epochs=100).load(
                self.index_path / 'ivf_ann_model')

        if (self.index_path / 'ivf_index').exists():
            self.ivf_index = IVFIndex().load(self.index_path / 'ivf_index')

    def _kmeans_clustering(self, data, rebuild=False):
        """kmeans clustering"""
        if rebuild:
            self.ann_model = self.ann_model.partial_fit(data)
            labels = self.ann_model.labels_
        else:
            raise_if(ValueError, not self.ann_model.fitted, "The model has not been fitted yet.")
            labels = self.ann_model.predict(data)

        return labels

    def _build_index(self, rebuild=False):
        """
        Build the IVF index more efficiently.
        """
        if rebuild:
            self.ivf_index = IVFIndex()

        filenames = self.storage_worker.get_all_files()
        already_indexed = self.ivf_index.all_external_ids()

        for filename in filenames:
            data, indices = self.dataloader(filename)

            isin = np.isin(indices, already_indexed)

            if isin.all():
                continue
            else:
                data = data[~isin]
                indices = indices[~isin]

            labels = self._kmeans_clustering(data, rebuild=rebuild)

            for idx, label in zip(indices, labels):
                self.ivf_index.add_entry(label, filename, idx)

    def _check_if_refit(self, n_clusters, always_refit=False):
        if self.ann_model is None or always_refit:
            self.ann_model = BatchKMeans(n_clusters=n_clusters, batch_size=10240, epochs=config.LYNSE_KMEANS_EPOCHS)

        REFIT = False

        if (self.ann_model.n_clusters != n_clusters or
                self.ann_model.epochs != config.LYNSE_KMEANS_EPOCHS or
                self.ann_model.batch_size != 10240):
            self.ann_model = BatchKMeans(n_clusters=n_clusters, batch_size=10240, epochs=config.LYNSE_KMEANS_EPOCHS)
            REFIT = True

        return REFIT

    def build_index(self, n_clusters=32, rebuild=False):
        """
        Build the index for clustering.

        Parameters:
            n_clusters (int): The number of clusters.
            rebuild (bool): Whether to rebuild the index.

        Returns:
            None
        """
        raise_if(ValueError, not isinstance(n_clusters, int) or n_clusters <= 0,
                 'n_clusters must be int and greater than 0')
        raise_if(TypeError, not isinstance(rebuild, bool), 'rebuild must be bool')

        all_partition_size = self.storage_worker.get_shape()[0]
        if all_partition_size < 100000:
            return

        with ThreadLock():
            REBUILD = self._check_if_refit(n_clusters, always_refit=rebuild) or rebuild

            if REBUILD:
                # delete the old index
                self.remove_index()
                self.ann_model = BatchKMeans(n_clusters=n_clusters, batch_size=10240, epochs=config.LYNSE_KMEANS_EPOCHS)
                self._build_index(rebuild=True)
            else:
                already_clustered_rows = len(self.ivf_index)

                if already_clustered_rows == 0:
                    self._build_index(rebuild=True)
                elif already_clustered_rows == all_partition_size:
                    return
                else:
                    if all_partition_size - already_clustered_rows >= 100000:
                        self._build_index(rebuild=True)
                    else:
                        self._build_index(rebuild=False)

            # save ivf index and k-means model
            self.ann_model.save(self.index_path / 'ivf_ann_model')

            self.ivf_index.save(self.index_path / 'ivf_index')

    def fast_insert(self, start_filename):
        """
        Insert the data into the index.

        Parameters:
            start_filename (str): The filename to start with.

        Returns:
            None
        """
        filenames = self.storage_worker.get_all_files()
        filenames = [filename for filename in filenames if int(filename.split('_')[-1]) >=
                     int(start_filename.split('_')[-1])]

        already_indexed = self.ivf_index.all_external_ids()

        only_run_once = False
        for filename in filenames:
            data, indices = self.dataloader(filename)

            if not only_run_once:
                isin = np.isin(indices, already_indexed)

                if isin.all():
                    continue
                else:
                    data = data[~isin]
                    indices = indices[~isin]

                only_run_once = True

            labels = self._kmeans_clustering(data, rebuild=True)

            for idx, label in zip(indices, labels):
                self.ivf_index.add_entry(label, filename, idx)

    def remove_index(self):
        """
        Remove the index.

        If all indices are removed, the index file will be removed.

        Returns:
            None
        """
        self.ann_model = None
        self.ivf_index.clear_all()
        if (self.index_path / 'ivf_ann_model').exists():
            (self.index_path / 'ivf_ann_model').unlink()
        if (self.index_path / 'ivf_index').exists():
            (self.index_path / 'ivf_index').unlink()

    def predict(self, data):
        """
        Predict the cluster label for each sample.

        Parameters:
            data (np.ndarray): The input data.

        Returns:
            np.ndarray: The predicted cluster labels.
        """
        return self.ann_model.predict(data)



from lynse.api.native_api.high_level import LocalClient
from lynse.api.native_api.low_level import ExclusiveDB
from lynse.core_components.fields_cache import Filter, FieldCondition, MatchField, MatchID
from lynse import VectorDBClient
from lynse.api.http_api.client_api import HTTPClient, Collection



from ..core_components.locks import ThreadLock


class SafeDict(dict):
    def __init__(self):
        super().__init__()
        self.lock = ThreadLock()

    def __setitem__(self, key, value):
        with self.lock:
            super().__setitem__(key, value)

    def __delitem__(self, key):
        with self.lock:
            super().__delitem__(key)

    def pop(self, key, default=None):
        with self.lock:
            if key not in self:
                return default
            return super().pop(key)

    def clear(self):
        with self.lock:
            super().clear()

    def update(self, other):
        with self.lock:
            super().update(other)



import re
import time
from functools import wraps
from pathlib import Path

import numpy as np


class OpsError(Exception):
    """An exception that is raised when an error occurs during a database operation."""

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class SearchResultsCache:
    """A decorator that caches the results of a function call with the same arguments."""

    def __init__(self, max_size=1000, expire_seconds=3600):
        from collections import OrderedDict

        self.cache = OrderedDict()
        self.max_size = max_size
        self.expire_seconds = expire_seconds

    def clear_cache(self):
        self.cache.clear()

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key_parts = []
            for arg in args[1:]:  # ignore the self parameter
                if hasattr(arg, 'tobytes'):
                    key_parts.append(('vector', arg.tobytes()))
                elif hasattr(arg, '__dict__'):
                    key_parts.append(arg.__dict__)
                elif isinstance(arg, list):
                    key_parts.append(tuple(arg))
                else:
                    key_parts.append(arg)

            for k, v in kwargs.items():
                if hasattr(v, 'tobytes'):
                    key_parts.append((k, v.tobytes()))
                elif hasattr(v, '__dict__'):
                    key_parts.append(v.__dict__)
                elif isinstance(v, list):
                    key_parts.append(tuple(v))
                else:
                    key_parts.append((k, v))

            key = tuple(key_parts)

            current_time = time.mktime(time.gmtime())
            if key in self.cache:
                result, timestamp = self.cache[key]
                if current_time - timestamp < self.expire_seconds:
                    return result

            result = func(*args, **kwargs)
            self.cache[key] = (result, current_time)

            if len(self.cache) > self.max_size:
                self.cache.popitem(last=False)

            return result

        wrapper.clear_cache = self.clear_cache
        return wrapper


def unavailable_if_deleted(func):
    """A decorator that detects if the function is called after the object is deleted."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if hasattr(args[0], '_search'):
            # self is the first parameter
            db_name = Path(args[0]._database_path).name

            if args[0]._matrix_serializer.IS_DELETED:
                raise OpsError(f"The collection `{db_name}` has been deleted, and the `{func.__name__}` function "
                               f"is unavailable.")
        else:
            db_name = Path(args[0].root_path).name

            if args[0].STATUS == 'DELETED':
                raise OpsError(f"The `{db_name}` has been deleted, and the `{func.__name__}` function "
                               f"is unavailable.")

        return func(*args, **kwargs)

    return wrapper


def unavailable_if_empty(func):
    """A decorator that detects if the database is empty."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if args[0].shape[0] == 0:
            raise OpsError(f"The database is empty, and the `{func.__name__}` function is unavailable.")

        return func(*args, **kwargs)

    return wrapper


def load_chunk_file(filename):
    np_array = np.asarray(np.load(filename, mmap_mode='r'))
    return np_array


class NpyLoader:
    def __init__(self, filename):
        self.filename = filename
        self.np_array = None
        self.mmap = None

    def __enter__(self):
        self.np_array = load_chunk_file(self.filename)

        self.mmap = getattr(self.np_array, 'base', None)

        return self.np_array
    
    def __exit__(self, exc_type, exc_value, traceback):
        if self.np_array is not None:
            del self.np_array
            if self.mmap is not None and hasattr(self.mmap, 'close'):
                self.mmap.close()



def drop_duplicated_substr(source, target) -> str:
    """
    Remove all occurrences of the target string from the source string.

    Parameters:
        source (str): The source string.
        target (str): The target string.

    Returns:
        str: The source string with all occurrences of the target string removed
    """
    t_len = len(target)
    s_len = len(source)

    indices_to_remove = []
    i = 0

    while i <= s_len - t_len:
        if source[i:i + t_len] == target:
            indices_to_remove.append(i)
            i += t_len
        else:
            i += 1

    result = []
    last_index = 0

    for start_index in indices_to_remove:
        result.append(source[last_index:start_index])
        last_index = start_index + t_len

    result.append(source[last_index:])

    return ''.join(result)


def find_first_file_with_substr(directory, substr):
    """
    Find the first file with the specified substring or wildcard in the directory.

    Parameters:
        directory (str or Pathlike): The directory to search.
        substr (str): The substring or wildcard pattern of the file to search for.

    Returns:
        path: The path to the first file with the specified substring or wildcard pattern in the directory.
    """
    # Convert wildcard pattern to regular expression
    regex_pattern = re.compile(re.escape(substr).replace(r'\*', '.*'))

    for file in Path(directory).iterdir():
        if regex_pattern.search(file.name):
            return file.absolute()

    return None


def collection_repr(collection):
    """
    Get the string representation of a collection.

    Parameters:
        collection (Collection): The collection to represent.

    Returns:
        str: The string representation of the collection.
    """
    return (f'{collection.name}CollectionInstance(\n'
            f'    database="{collection._database_name}", \n'
            f'    collection="{collection._collection_name}", \n'
            f'    shape={collection.shape}'
            f'\n)')


def sort_and_get_top_k(arr, k):
    """
    Sort each row of a 1D or 2D array and return the top-k indices and values for each row.

    Parameters:
        arr (np.ndarray): 1D or 2D numpy array
        k (int): Number of top elements to return

    Returns:
        tuple: (top_k_indices, top_k_values)
            top_k_indices: Indices of top-k elements for each row
            top_k_values: Values of top-k elements for each row
    """
    arr = np.asarray(arr)
    
    if arr.ndim == 1:
        arr = arr.reshape(1, -1)
    
    n_rows, n_cols = arr.shape
    
    k = min(k, n_cols)
    
    sorted_indices = np.argsort(arr, axis=1)[:, ::-1]
    
    top_k_indices = sorted_indices[:, :k]
    top_k_values = np.take_along_axis(arr, top_k_indices, axis=1)
    
    return top_k_indices, top_k_values



import os
from typing import List

from .filter import MatchField, MatchID, FieldCondition, Filter, MatchRange
from .expression_parse import ExpressionParser


class IndexSchema:
    def __init__(self):
        self.indices = {}

    @staticmethod
    def _check_field_name(field_name: str):
        if ':id:' == field_name.strip():
            raise ValueError("The field name ':id:' is reserved.")

    def _add_index(self, field_name: str, field_type):
        self._check_field_name(field_name)
        self.indices[field_name] = field_type

    def add_string_index(self, field_name: str):
        self._add_index(field_name, str)

    def add_int_index(self, field_name: str):
        self._add_index(field_name, int)

    def to_dict(self):
        indices = {k: v.__name__ for k, v in self.indices.items()}
        return indices

    def load_from_dict(self, schema_dict):
        self.indices = {k: eval(v) for k, v in schema_dict.items()}
        return self


class FieldsCache:
    def __init__(self, filepath=None, as_temp_file=False):
        """
        Create a FieldsCache instance.

        Parameters:
            filepath: str
                The path to the file where the cache will be stored.
            as_temp_file: bool
                If True, the cache will be stored in a temporary file.
        """
        from ...core_components.fields_cache.fields_query import FieldsQuery
        from ...core_components.fields_cache.fields_storage import FieldsStorage

        self.storage = FieldsStorage(filepath, as_temp_file=as_temp_file)
        self.query_handler = FieldsQuery(self.storage)
        self.filepath = filepath

    def store(self, data: dict, external_id: int):
        """
        Store a record in the cache.

        Parameters:
            data: dict
                The record to store.
            external_id: int
                The external ID of the record.

        Returns:
            int: The internal ID of the record.
        """
        return self.storage.store(data, external_id)

    def query(self, query_filter, filter_ids=None, return_ids_only=True):
        """
        Query the fields cache.

        Parameters:
            query_filter: Filter or dict or FieldExpression (string)
                The filter object or the specify data to filter.
            filter_ids: List[int]
                The list of external IDs to filter.
            return_ids_only: bool
                If True, only the external IDs will be returned.

        Returns:
            List[dict]: The records. If not return_ids_only, the records will be returned.
            List[int]: The external IDs. If return_ids_only, the external IDs will be returned.
        """
        if (not isinstance(query_filter, Filter) and not isinstance(query_filter, dict)
                and not isinstance(query_filter, str)):
            raise ValueError("The filter_instance must be an instance of Filter or a dict or a FieldExpression string.")

        return self.query_handler.query(query_filter, filter_ids, return_ids_only)

    def retrieve(self, external_id, include_external_id=False):
        """
        Retrieve a record from the cache.

        Parameters:
            external_id: int
                The external ID of the record.
            include_external_id: bool
                If True, the external ID will be included in the record.

        Returns:
            dict: The record.
        """
        return self.query_handler.retrieve(external_id, include_external_id=include_external_id)

    def retrieve_ids(self, external_ids: List[int], include_external_id: bool = False):
        """
        Retrieve records from the cache.

        Parameters:
            external_ids: List[int]
                The external IDs of the records.
            include_external_id: bool
                If True, the external IDs will be included in the records.

        Returns:
            List[dict]: The records.
        """
        return self.query_handler.retrieve_ids(external_ids, include_external_id=include_external_id)

    def concat(self, other: 'FieldsCache') -> 'FieldsCache':
        """
        Concatenate two caches.

        Parameters:
            other: FieldsCache
                The other cache to concatenate.

        Returns:
            FieldsCache: The concatenated cache.
        """
        if not isinstance(other, FieldsCache):
            raise ValueError("The other cache must be an instance of FieldsCache.")
        for external_id, data in other.storage.retrieve_all():
            self.store(data, external_id)
        return self

    def delete(self):
        """
        Delete the cache.
        """
        self.storage.delete()

        if hasattr(self, 'query_handler'):
            del self.query_handler

    def commit(self):
        """
        Commit the cache.

        This method is used to save the cache to disk.
        """
        self.storage.commit()

    def build_index(self, schema: IndexSchema, rebuild_if_exists=False):
        """
        Build an index for the cache.

        Parameters:
            schema (IndexSchema or Field name string): The index schema or the field name string. 
                When passing the field name string, the field name must be wrapped with ':',
                like ':vector:', ':timestamp:'.
            rebuild_if_exists: bool
                If True, the index will be rebuilt if it already exists.

        Returns:
            None

        Note:
            The :id: is a reserved field name and cannot be used.
        """
        if isinstance(schema, str):
            if schema == ':id:':
                raise ValueError("The field name ':id:' is reserved.")
            
            schema_type = self.storage.list_fields().get(schema, None)
            schema = schema.strip(':')

            if schema_type is None:
                raise ValueError(f"Field '{schema}' not found in the cache.")

            schema_cls = IndexSchema()

            if schema_type == 'str':
                schema_cls.add_string_index(schema)
            elif schema_type == 'int':
                schema_cls.add_int_index(schema)
            else:
                raise ValueError(f"Unsupported field type: {schema_type}")

            schema = schema_cls

        if not schema.indices:
            raise ValueError("The index schema is empty.")
        self.storage.build_index(schema.indices, rebuild_if_exists)
        self.storage.index.save(self.storage.index_path)

    def remove_index(self, field_name: str):
        """
        Remove an index from the cache.

        Parameters:
            field_name: str
                The name of the field.

        Returns:
            None
        """
        self.storage.remove_index(field_name)
        if not self.storage.index.indices:
            if os.path.exists(self.storage.index_path):
                os.remove(self.storage.index_path)
        else:
            self.storage.index.save(self.storage.index_path)

    def remove_all_field_indices(self):
        """
        Remove all the field indices from the cache.

        Returns:
            None
        """
        indices = self.list_indices()
        for index in indices:
            self.remove_index(index)

    def list_indices(self):
        """
        List the indices in the cache.

        Returns:
            List[str]: The list of indices.
        """
        return [":" + i + ":" if i not in [':id:', ''] else i 
                for i in list(self.storage.index.indices.keys())]



import re
import operator
from abc import ABC, abstractmethod

from .filter import MatchField, MatchID, FieldCondition, Filter, MatchRange

operator_map = {
    '==': operator.eq,
    '!=': operator.ne,
    '>': operator.gt,
    '>=': operator.ge,
    '<': operator.lt,
    '<=': operator.le,
    'in': operator.eq,
    'not in': operator.eq
}


class ASTNode(ABC):
    @abstractmethod
    def evaluate(self):
        """
        Evaluates the AST node.

        Returns:
            Filter: The evaluated filter.
        """
        pass


class BinaryOpNode(ASTNode):
    def __init__(self, left, op, right):
        self.left = left
        self.op = op
        self.right = right

    def evaluate(self):
        """
        Evaluates the binary operation node.

        Returns:
            Filter: A filter based on the operator combination.
        """
        left_val = self.left.evaluate()
        right_val = self.right.evaluate()
        if self.op == 'and':
            return Filter(must=left_val.must_fields + right_val.must_fields,
                          any=left_val.any_fields + right_val.any_fields,
                          must_not=left_val.must_not_fields + right_val.must_not_fields)
        elif self.op == 'or':
            return Filter(any=left_val.must_fields + right_val.must_fields + left_val.any_fields + right_val.any_fields)
        else:
            raise ValueError(f"Unknown operator: {self.op}")


class UnaryOpNode(ASTNode):
    def __init__(self, op, expr):
        self.op = op
        self.expr = expr

    def evaluate(self):
        """
        Evaluates the unary operation node.

        Returns:
            Filter: For the 'not' operation, returns the inverted filter.
        """
        result = self.expr.evaluate()
        if self.op == 'not':
            return Filter(
                must=result.must_not_fields,
                must_not=result.must_fields,
                any=result.any_fields
            )
        else:
            raise ValueError(f"Unknown unary operator: {self.op}")


class ConditionNode(ASTNode):
    def __init__(self, field, op, value):
        self.field = field
        self.op = op
        self.value = value

    def evaluate(self):
        """
        Evaluates the condition node.

        Returns:
            Filter: A filter based on the condition.
        """
        field = self.field.replace(':', '')
        field = ':id:' if field == 'id' else field

        if field == ':id:':
            value = [int(v.strip()) for v in self.value.strip('[]').split(',')]
            if self.op == 'not in':
                return Filter(must_not=[FieldCondition(field, MatchID(value))])
            elif self.op == 'in':
                return Filter(must=[FieldCondition(field, MatchID(value))])

        if self.op in ['in', 'not in']:
            value = [v.strip() for v in self.value.strip('[]').split(',')]
            value = [self.parse_number(v) for v in value]
        else:
            value = self.parse_number(self.value)
        comparator = operator_map[self.op]
        condition = FieldCondition(field,
                                   MatchField(value, comparator, all_comparators=False, not_in=self.op == 'not in'))
        if self.op.startswith('not') and self.op != 'not in':
            return Filter(must_not=[condition])
        else:
            return Filter(must=[condition])

    @staticmethod
    def parse_number(value):
        """
        Parses a string into a number (int or float) if possible.

        Parameters:
            value (str): The string to be parsed.

        Returns:
            int, float, or str: The parsed value.
        """
        try:
            if '.' in value:
                return float(value)
            return int(value)
        except ValueError:
            return value.strip("'\"")


class RangeNode(ASTNode):
    def __init__(self, field, start, end, start_op, end_op):
        self.field = field
        self.start = start
        self.end = end
        self.start_op = start_op
        self.end_op = end_op

    def evaluate(self):
        """
        Evaluates the range node.

        Returns:
            Filter: A filter based on the range condition.
        """
        field = self.field.replace(':', '')
        field = ':id:' if field == 'id' else field

        start = self.parse_number(self.start)
        end = self.parse_number(self.end)

        # Determine actual start and end values
        if self.start_op in ['>', '>='] and self.end_op in ['>', '>=']:
            start, end = end, start

        if start > end:
            raise ValueError(f"Invalid range expression: Start value {start} is greater than end value {end}.")

        start_inclusive = self.end_op in ['>=', '<=']
        end_inclusive = self.start_op in ['<=', '>=']

        if start == end:
            if not (start_inclusive and end_inclusive):
                raise ValueError(f"Invalid range expression: When the start value equals the end value, "
                                 f"both must use one of ['==', '!=', '>=', '<='].")
            inclusive = True
        else:
            if start_inclusive and end_inclusive:
                inclusive = True
            elif not start_inclusive and not end_inclusive:
                inclusive = False
            elif start_inclusive:
                inclusive = 'left'
            else:
                inclusive = 'right'

        condition = FieldCondition(field, MatchRange(start, end, inclusive=inclusive))
        return Filter(must=[condition])

    @staticmethod
    def parse_number(value):
        """
        Parses a string into a number (int or float) if possible.

        Parameters:
            value (str): The string to be parsed.

        Returns:
            int, float, or str: The parsed value.
        """
        try:
            if '.' in value:
                return float(value)
            return int(value)
        except ValueError:
            return value.replace(':', '')


class ExpressionParser:
    """
    FieldExpression is a string parser used to parse field expressions. 

    It can reassemble the parsing result into a Filter object of LynseDB, which can filter the query result. 
    Its main implementation idea is to efficiently create Filters 
        without requiring users to learn the use of complex filtering components such as MatchRange and FieldCondition.

    It is an important component of LynseDB and can effectively improve the user-friendliness of LynseDB.

    Supported syntax:
        - :field: represents a field, e.g., :order:, :name:, etc.
        - :id: special reserved word for matching IDs

    Supported comparison operators:
        - ==
        - !=
        - \>=
        - <=
        - \>
        - <

    Supported logical operators:
        - and
        - or
        - not
        - in

    Supported parentheses:
        - ()
        
    Examples:
        - :order: > 1
        - :order: >= 1
        - :order: >= 1 and :name: >= 1.3
        - not (:order: == 2)
        - :order: >= 1 and :name: != 1.3 and (:order: == 2 or :name: == 3) and not (:tt: == 2)
        - :id: in [1, 2]
        - :id: not in [1, 2]
        - :id: in [1, 2] and :order: >= 1 and :name: >= 1.3 and (:order: == 2 or :name: == 3) and not (:tt: == 2)

    Usage:
        >>> expression = ":order: > 1 and :name: >= 1.3 and (:order: == 2 or :name: == 3) and not (:tt: == 2)"
        >>> parser = ExpressionParser(expression)
        >>> pfilter = parser.to_filter()
        >>> print(pfilter.to_dict())
        {'must_fields': [
                {'key': 'order', 'matcher': {'value': 1, 'comparator': 'gt'}},
                {'key': 'name', 'matcher': {'value': 1.3, 'comparator': 'ge'}}
            ],
        'any_fields': [
                {'key': 'order', 'matcher': {'value': 2, 'comparator': 'eq'}},
                {'key': 'name', 'matcher': {'value': 3, 'comparator': 'eq'}}
            ],
        'must_not_fields': [
                {'key': 'tt', 'matcher': {'value': 2, 'comparator': 'eq'}}
            ]
        }

    Efficient syntax:
        - Use () to group expressions for better readability and maintainability.
        Specially when using `not` operator.
        - Use `not in` instead of `not :field: in [value]` when possible.
        - Use `not (:field: in [value])` or `not (:field: in [value])`, it can be more efficient.


    """

    def __init__(self, expression):
        """
        Initialize ExpressionParser.

        Parameters:
            expression (str): The expression to be parsed.
        """
        if not expression or expression.strip() == '':
            raise ValueError("Expression cannot be empty.")
        self.expression = expression
        self.tokens = self.tokenize(expression)
        self.current = 0

        if not self.tokens:
            raise ValueError("Unable to parse expression, please check the syntax.")

    @staticmethod
    def tokenize(expression):
        """
        Tokenize the input expression.

        Parameters:
            expression (str): The expression to be tokenized.

        Returns:
            list: A list of tokens.
        """
        token_pattern = re.compile(
            r"(:\w+:|==|!=|>=|<=|>|<|in|not in|and|or|not|\(|\)|\[.*?\]|\d+\.\d+|\d+|'[^']*'|\"[^\"]*\")")
        return token_pattern.findall(expression)

    def parse(self):
        """
        Parse the expression.

        Returns:
            ASTNode: The root node of the abstract syntax tree.
        """
        return self.parse_or()

    def parse_or(self):
        """
        Parse 'or' expression.

        Returns:
            ASTNode: The parsed node.
        """
        node = self.parse_and()
        while self.current < len(self.tokens) and self.tokens[self.current] == 'or':
            self.current += 1
            right = self.parse_and()
            node = BinaryOpNode(node, 'or', right)
        return node

    def parse_and(self):
        """
        Parse 'and' expression.

        Returns:
            ASTNode: The parsed node.
        """
        node = self.parse_not()
        while self.current < len(self.tokens) and self.tokens[self.current] == 'and':
            self.current += 1
            right = self.parse_not()
            node = BinaryOpNode(node, 'and', right)
        return node

    def parse_not(self):
        """
        Parse 'not' expression.

        Returns:
            ASTNode: The parsed node.
        """
        if self.current < len(self.tokens) and self.tokens[self.current] == 'not':
            self.current += 1
            expr = self.parse_not()
            return UnaryOpNode('not', expr)
        else:
            return self.parse_primary()

    def parse_primary(self):
        """
        Parse the basic expression (parentheses and conditions).

        Returns:
            ASTNode: The parsed node.
        """
        if self.tokens[self.current] == '(':
            self.current += 1
            node = self.parse_or()
            if self.current >= len(self.tokens) or self.tokens[self.current] != ')':
                raise ValueError("Missing right parenthesis")
            self.current += 1 
            return node
        else:
            return self.parse_condition()

    @staticmethod
    def parse_number(value):
        """
        Parse a string into a number (int or float) if possible.

        Parameters:
            value (str): The string to be parsed.

        Returns:
            int, float, or str: The parsed value.
        """
        try:
            return int(value) if value.isdigit() else float(value)
        except ValueError:
            return value

    def compare_values(self, op1, val1, op2, val2):
        """
        Compare two values and return the stricter limit.

        Parameters:
            op1 (str): The operator of the first value.
            val1 (str): The first value.
            op2 (str): The operator of the second value.
            val2 (str): The second value.
        """
        num1 = self.parse_number(val1)
        num2 = self.parse_number(val2)

        if op1 in ['>', '>='] and op2 in ['<', '<=']:
            if num1 > num2:
                return op2, val2  # The stricter upper limit
            else:
                return op1, val1  # The stricter lower limit
        elif op1 in ['<', '<='] and op2 in ['>', '>=']:
            if num1 < num2:
                return op2, val2  # The stricter lower limit
            else:
                return op1, val1  # The stricter upper limit
        else:
            # For the same direction inequality, return the stricter limit
            if op1 in ['>', '>=']:
                return (op1, val1) if num1 > num2 else (op2, val2)
            else:
                return (op1, val1) if num1 < num2 else (op2, val2)

    def parse_condition(self):
        """
        Parse conditions and ranges.

        Returns:
            ASTNode: The parsed node (ConditionNode or RangeNode).
        """
        if self.current < len(self.tokens) - 2 and self.tokens[self.current + 1] in ['>', '<', '>=', '<=']:
            # Detect range expressions, e.g., "1 < :field: < 10"
            left = self.tokens[self.current]
            op1 = self.tokens[self.current + 1]
            middle = self.tokens[self.current + 2]
            if self.current + 3 < len(self.tokens) and self.tokens[self.current + 3] in ['>', '<', '>=', '<=']:
                op2 = self.tokens[self.current + 3]
                right = self.tokens[self.current + 4]
                self.current += 5
                return RangeNode(middle, left, right, op1, op2)
        # Standard condition parsing
        left = self.tokens[self.current]
        self.current += 1
        if self.current >= len(self.tokens):
            raise ValueError("Expression is incomplete, missing operator and value.")
        op = self.tokens[self.current]
        if op == 'not' and self.current + 1 < len(self.tokens):
            op = 'not ' + self.tokens[self.current + 1]
            self.current += 1
        self.current += 1
        if self.current >= len(self.tokens):
            raise ValueError("Expression is incomplete, missing value.")
        right = self.tokens[self.current]
        self.current += 1

        # Check if the field is on the left side
        if ':' in left:
            return ConditionNode(left, op, right)
        else:
            # Field on the right side, reverse the operator
            op = self.reverse_op(op)
            return ConditionNode(right, op, left)

    @staticmethod
    def reverse_op(op):
        """
        Reverse the comparison operator.

        Parameters:
            op (str): The operator to be reversed.

        Returns:
            str: The reversed operator.
        """
        reverse_map = {'>': '<', '<': '>', '>=': '<=', '<=': '>='}
        return reverse_map.get(op, op)

    def to_filter(self):
        """
        Get the final filter by parsing and evaluating the expression.

        Returns:
            Filter: The generated filter.
        """
        ast = self.parse()
        return ast.evaluate()



import operator

import msgpack
from spinesUtils.asserts import raise_if


class MatchRange:
    name: str = 'MatchRange'

    def __init__(self, start, end, inclusive=True):
        """
        Initialize a MatchRange instance with a start and end value.

        Parameters:
            start (any): The start value of the range.
            end (any): The end value of the range.
            inclusive (bool or str): Whether the range is inclusive of the start and end values.
                                     Can be True, False, "left", or "right".
        """
        self.start = start
        self.end = end
        self.inclusive = inclusive

    def match(self, data_value):
        """
        Check if the data_value falls within the range.

        Parameters:
            data_value (any): The value to compare against the range.

        Returns:
            bool: True if data_value is within the range, otherwise False.
        """
        if self.inclusive is True:
            return self.start <= data_value <= self.end
        elif self.inclusive == "left":
            return self.start <= data_value < self.end
        elif self.inclusive == "right":
            return self.start < data_value <= self.end
        else:
            return self.start < data_value < self.end


class MatchField:
    name: str = 'MatchField'

    def __init__(self, value, comparator=operator.eq, all_comparators=False, not_in=False):
        """
        Initialize a MatchField instance with a specific value and comparator.

        Parameters:
            value (list or tuple or any object that implements comparison functions such as __ eq__): The value to compare the data attribute with.
            comparator: The comparator function to apply to the data attribute.
            all_comparators: Whether to apply the comparator to all values in the list or tuple.
                If True, all values in the list or tuple must satisfy the comparison condition.
                If False, at least one value in the list or tuple must satisfy the comparison condition.
            not_in: Whether to negate the comparison condition.
        """
        self.value = value
        self.comparator = comparator
        self.all_comparators = all_comparators
        self.not_in = not_in
        raise_if(ValueError, all_comparators is True and not_in is True,
                 "all_comparators and not_in cannot be True at the same time.")

    def match(self, data_value):
        """
        Apply the comparator to compare the data attribute with the specified value.

        Parameters:
            data_value: The value of the data attribute to compare.

        Returns:
            bool: True if the data attribute matches the specified value, otherwise False.
        """
        if isinstance(self.value, (list, tuple)):
            if not self.not_in:
                if self.all_comparators:
                    return all([self.comparator(data_value, value) for value in self.value])
                else:
                    return data_value in self.value
            else:
                return data_value not in self.value

        return self.comparator(data_value, self.value)


class MatchID:
    name: str = 'MatchID'

    def __init__(self, ids: list):
        """
        Initialize an MatchID instance.

        Parameters:
            ids (list): The indices to filter the numpy array.
        """
        raise_if(TypeError, not isinstance(ids, list), "Invalid ids type.")

        self.indices = ids

    def match(self, external_id):
        """
        Filter the numpy array according to the specified indices.

        Parameters:
            external_id (int): The external id to filter.

        Returns:
            bool: True if the external id is in the specified indices, otherwise False.
        """
        return external_id in self.indices


class FieldCondition:
    def __init__(self, key=None, matcher=None):
        """
        Initialize a FieldCondition instance with a specific key and matcher.

        Parameters:
            key: The key of the data attribute to compare.
                If using MatchID, this should be the special key ':id:'.
            matcher: The MatchField, MatchID, or MatchRange instance to compare the data attribute with.
        """
        self.matcher = matcher
        if matcher is not None:
            self.key = ':id:' if isinstance(matcher, MatchID) else key
        else:
            self.key = key

        raise_if(TypeError, not isinstance(self.matcher, (MatchField, MatchID, MatchRange)), "Invalid matcher type.")

    def evaluate(self, data: dict, external_id: int):
        """
        Evaluate the condition against a given dictionary.

        Parameters:
            data: A dictionary with the key specified in the condition.
            external_id: The external id to filter.

        Returns:
            bool: True if the data attribute matches the specified value, otherwise False.
        """
        if self.key == ':id:':
            attribute_value = external_id
        else:
            attribute_value = data.get(self.key)

        if attribute_value is not None:
            return self.matcher.match(attribute_value)
        return False
    

class Filter:
    __slots__ = ['must_fields', 'any_fields', 'must_not_fields']

    def __init__(self, must=None, any=None, must_not=None):
        """
        Initialize a Filter instance with must and any conditions.

        The filter result must satisfy both must_fields and any_fields, but not must_not_fields.

        Parameters:
            must: A list of conditions that must be satisfied.
            any: A list of conditions where at least one must be satisfied.
            must_not: A list of conditions that must not be satisfied.
        """
        self.must_fields = []
        self.any_fields = []
        self.must_not_fields = []

        if must:
            for condition in must:
                if isinstance(condition, FieldCondition):
                    self.must_fields.append(condition)
                else:
                    raise ValueError("Invalid condition type.")

        if any:
            for condition in any:
                if isinstance(condition, FieldCondition):
                    self.any_fields.append(condition)
                else:
                    raise ValueError("Invalid condition type.")

        if must_not:
            for condition in must_not:
                if isinstance(condition, FieldCondition):
                    if isinstance(getattr(condition, "matcher"), MatchField):
                        if condition.matcher.not_in:
                            condition.matcher.not_in = False  # Needs to reverse the not_in flag in must_not conditions
                            condition.matcher.all_comparators = False
                    self.must_not_fields.append(condition)
                else:
                    raise ValueError("Invalid condition type.")

        self.must_fields = self._to_unique_list(self.must_fields)
        self.any_fields = self._to_unique_list(self.any_fields)
        self.must_not_fields = self._to_unique_list(self.must_not_fields)

    def to_dict(self):
        """
        Convert the filter to a dictionary.

        Returns:
            dict: A dictionary representation of the filter.
        """
        return {
            'must_fields': [
                {
                    'ids': condition.matcher.indices
                } if isinstance(condition.matcher, MatchID) else {
                    'key': condition.key,
                    'matcher': {
                        'value': condition.matcher.value,
                        'comparator': condition.matcher.comparator.__name__,
                        'all_comparators': condition.matcher.all_comparators,
                        'not_in': condition.matcher.not_in
                    }
                } if isinstance(condition.matcher, MatchField) else {
                    'key': condition.key,
                    'matcher': {
                        'start': condition.matcher.start,
                        'end': condition.matcher.end,
                        'inclusive': condition.matcher.inclusive
                    }
                } for condition in self.must_fields] if self.must_fields else [],
            'any_fields': [
                {
                    'ids': condition.matcher.indices
                } if isinstance(condition.matcher, MatchID) else {
                    'key': condition.key,
                    'matcher': {
                        'value': condition.matcher.value,
                        'comparator': condition.matcher.comparator.__name__,
                        'all_comparators': condition.matcher.all_comparators,
                        'not_in': condition.matcher.not_in
                    }
                } if isinstance(condition.matcher, MatchField) else {
                    'key': condition.key,
                    'matcher': {
                        'start': condition.matcher.start,
                        'end': condition.matcher.end,
                        'inclusive': condition.matcher.inclusive
                    }
                } for condition in self.any_fields] if self.any_fields else [],
            'must_not_fields': [
                {
                    'ids': condition.matcher.indices
                } if isinstance(condition.matcher, MatchID) else {
                    'key': condition.key,
                    'matcher': {
                        'value': condition.matcher.value,
                        'comparator': condition.matcher.comparator.__name__,
                        'all_comparators': condition.matcher.all_comparators,
                        'not_in': condition.matcher.not_in
                    }
                } if isinstance(condition.matcher, MatchField) else {
                    'key': condition.key,
                    'matcher': {
                        'start': condition.matcher.start,
                        'end': condition.matcher.end,
                        'inclusive': condition.matcher.inclusive
                    }
                } for condition in self.must_not_fields] if self.must_not_fields else []
        }

    def load_dict(self, data):
        """
        Load the filter from a dictionary.

        Parameters:
            data: The dictionary to load the filter from.
        """
        self.must_fields = []
        self.any_fields = []
        self.must_not_fields = []

        for condition in data.get('must_fields', []):
            if 'ids' in condition:
                self.must_fields.append(FieldCondition(':id:', MatchID(condition['ids'])))
            elif 'value' in condition['matcher']:
                self.must_fields.append(FieldCondition(condition['key'],
                                                       MatchField(
                                                           condition['matcher']['value'],
                                                           getattr(operator, condition['matcher']['comparator']),
                                                           condition['matcher']['all_comparators'],
                                                           condition['matcher']['not_in']
                                                       )))
            else:
                self.must_fields.append(FieldCondition(condition['key'],
                                                       MatchRange(
                                                           condition['matcher']['start'],
                                                           condition['matcher']['end'],
                                                           condition['matcher']['inclusive'])))

        for condition in data.get('any_fields', []):
            if 'ids' in condition:
                self.any_fields.append(FieldCondition(':id:', MatchID(condition['ids'])))
            elif 'value' in condition['matcher']:
                self.any_fields.append(FieldCondition(condition['key'],
                                                      MatchField(
                                                          condition['matcher']['value'],
                                                          getattr(operator, condition['matcher']['comparator']),
                                                          condition['matcher']['all_comparators'],
                                                          condition['matcher']['not_in']
                                                      )))
            else:
                self.any_fields.append(FieldCondition(condition['key'],
                                                      MatchRange(
                                                          condition['matcher']['start'],
                                                          condition['matcher']['end'],
                                                          condition['matcher']['inclusive'])))

        for condition in data.get('must_not_fields', []):
            if 'ids' in condition:
                self.must_not_fields.append(FieldCondition(':id:', MatchID(condition['ids'])))
            elif 'value' in condition['matcher']:
                self.must_not_fields.append(FieldCondition(
                    condition['key'], MatchField(
                        condition['matcher']['value'],
                        getattr(operator, condition['matcher']['comparator']),
                        condition['matcher']['all_comparators'],
                        condition['matcher']['not_in']
                    )))
            else:
                self.must_not_fields.append(FieldCondition(
                    condition['key'], MatchRange(
                        condition['matcher']['start'],
                        condition['matcher']['end'],
                        condition['matcher']['inclusive'])))
        
        self.must_fields = self._to_unique_list(self.must_fields)
        self.any_fields = self._to_unique_list(self.any_fields)
        self.must_not_fields = self._to_unique_list(self.must_not_fields)
        
        return self

    @staticmethod
    def _to_unique_list(conditions):
        """
        Remove duplicate conditions from the list.

        Parameters:
            conditions: The list of conditions to remove duplicates from.

        Returns:
            list: A list of conditions with duplicates removed.
        """
        unique_conditions = []
        for condition in conditions:
            if not any(Filter._conditions_equal(condition, unique_cond) for unique_cond in unique_conditions):
                unique_conditions.append(condition)
        return unique_conditions

    @staticmethod
    def _conditions_equal(cond1, cond2):
        """
        Compare two FieldCondition objects for equality.

        Parameters:
            cond1: The first FieldCondition object.
            cond2: The second FieldCondition object.

        Returns:
            bool: True if the conditions are equal, False otherwise.
        """
        if not isinstance(cond1, FieldCondition) or not isinstance(cond2, FieldCondition):
            return False
        if cond1.key != cond2.key:
            return False
        if type(cond1.matcher) != type(cond2.matcher):
            return False
        if isinstance(cond1.matcher, (MatchField, MatchID, MatchRange)):
            return cond1.matcher.__dict__ == cond2.matcher.__dict__
        return cond1.matcher == cond2.matcher
    
    def __hash__(self):
        return hash(msgpack.packb(self.to_dict()))



import os
from copy import deepcopy
from pathlib import Path

from spinesUtils.asserts import raise_if, augmented_isinstance
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap


class Config:
    def __init__(self):
        self._LYNSE_LOG_LEVEL = self.get_env_variable('LYNSE_LOG_LEVEL', 'INFO', str, [str])
        self._LYNSE_LOG_PATH = self.get_env_variable('LYNSE_LOG_PATH', None, str, [str, None])
        self._LYNSE_TRUNCATE_LOG = self.get_env_variable('LYNSE_TRUNCATE_LOG', True, bool, [bool])
        self._LYNSE_LOG_WITH_TIME = self.get_env_variable('LYNSE_LOG_WITH_TIME', True, bool, [bool])
        self._LYNSE_KMEANS_EPOCHS = self.get_env_variable('LYNSE_KMEANS_EPOCHS', 100, int, [int])
        self._LYNSE_SEARCH_CACHE_SIZE = self.get_env_variable('LYNSE_SEARCH_CACHE_SIZE', 10000, int, [int])
        self._LYNSE_DEFAULT_ROOT_PATH = self.get_env_variable(
            'LYNSE_DEFAULT_ROOT_PATH',
            Path(os.path.expanduser('~/.LynseDB/databases/')),
            Path, [str, None, Path]
        )
        self._LYNSE_SEARCH_CACHE_EXPIRE_SECONDS = self.get_env_variable('LYNSE_SEARCH_CACHE_EXPIRE_SECONDS',
                                                                        3600, int, [int])

    @staticmethod
    def get_env_variable(name, default=None, default_type=str, type_allow_list=None):
        def type_cast(value):
            if value == 'None':
                return None
            if value == default:
                return default
            if default_type == bool and isinstance(name, str):
                return value.lower() == 'true'
            if default_type == str:
                return value
            else:
                try:
                    return default_type(value)  # will raise Exception if None
                except Exception:
                    return default  # include None

        if default is None:
            value = type_cast(os.environ.get(name))
            if type_allow_list is not None:
                raise_if(ValueError, not augmented_isinstance(value, tuple(type_allow_list)),
                         f"{name} must be in {type_allow_list}")
            return value
        else:
            return type_cast(os.environ.get(name, default))

    @property
    def LYNSE_LOG_LEVEL(self):
        """Log level"""
        return self._LYNSE_LOG_LEVEL

    @LYNSE_LOG_LEVEL.setter
    def LYNSE_LOG_LEVEL(self, value):
        self._LYNSE_LOG_LEVEL = value

    @property
    def LYNSE_LOG_PATH(self):
        """Log path"""
        return self._LYNSE_LOG_PATH

    @LYNSE_LOG_PATH.setter
    def LYNSE_LOG_PATH(self, value):
        self._LYNSE_LOG_PATH = value

    @property
    def LYNSE_TRUNCATE_LOG(self):
        """Whether to truncate log"""
        return self._LYNSE_TRUNCATE_LOG

    @LYNSE_TRUNCATE_LOG.setter
    def LYNSE_TRUNCATE_LOG(self, value):
        self._LYNSE_TRUNCATE_LOG = value

    @property
    def LYNSE_LOG_WITH_TIME(self):
        """Whether to include time in log"""
        return self._LYNSE_LOG_WITH_TIME

    @LYNSE_LOG_WITH_TIME.setter
    def LYNSE_LOG_WITH_TIME(self, value):
        self._LYNSE_LOG_WITH_TIME = value

    @property
    def LYNSE_KMEANS_EPOCHS(self):
        """Number of KMeans epochs"""
        return self._LYNSE_KMEANS_EPOCHS

    @LYNSE_KMEANS_EPOCHS.setter
    def LYNSE_KMEANS_EPOCHS(self, value):
        self._LYNSE_KMEANS_EPOCHS = value

    @property
    def LYNSE_SEARCH_CACHE_SIZE(self):
        """Search cache size"""
        return self._LYNSE_SEARCH_CACHE_SIZE

    @LYNSE_SEARCH_CACHE_SIZE.setter
    def LYNSE_SEARCH_CACHE_SIZE(self, value):
        self._LYNSE_SEARCH_CACHE_SIZE = value

    @property
    def LYNSE_SEARCH_CACHE_EXPIRE_SECONDS(self):
        """Search cache expire time in seconds"""
        return self._LYNSE_SEARCH_CACHE_EXPIRE_SECONDS

    @LYNSE_SEARCH_CACHE_EXPIRE_SECONDS.setter
    def LYNSE_SEARCH_CACHE_EXPIRE_SECONDS(self, value):
        self._LYNSE_SEARCH_CACHE_EXPIRE_SECONDS = value

    @property
    def LYNSE_DEFAULT_ROOT_PATH(self):
        """Default root path"""
        return self._LYNSE_DEFAULT_ROOT_PATH

    @LYNSE_DEFAULT_ROOT_PATH.setter
    def LYNSE_DEFAULT_ROOT_PATH(self, value):
        self._LYNSE_DEFAULT_ROOT_PATH = value

    def get_all_configs(self):
        return {
            'LYNSE_LOG_LEVEL': self.LYNSE_LOG_LEVEL,
            'LYNSE_LOG_PATH': self.LYNSE_LOG_PATH,
            'LYNSE_TRUNCATE_LOG': self.LYNSE_TRUNCATE_LOG,
            'LYNSE_LOG_WITH_TIME': self.LYNSE_LOG_WITH_TIME,
            'LYNSE_KMEANS_EPOCHS': self.LYNSE_KMEANS_EPOCHS,
            'LYNSE_SEARCH_CACHE_SIZE': self.LYNSE_SEARCH_CACHE_SIZE,
            'LYNSE_DEFAULT_ROOT_PATH': str(self.LYNSE_DEFAULT_ROOT_PATH),
            'LYNSE_SEARCH_CACHE_EXPIRE_SECONDS': self.LYNSE_SEARCH_CACHE_EXPIRE_SECONDS
        }


def generate_config_file(regenerate=False):
    config = Config()
    user_path = Path(os.path.expanduser('~'))
    config_path = user_path / '.lynsedb_configs.yaml'

    config_dict = deepcopy(config.get_all_configs())
    yaml = YAML()
    yaml.default_flow_style = False

    commented_config_dict = CommentedMap()

    for k, v in config_dict.items():
        if isinstance(v, Path):
            v = str(v)
        commented_config_dict[k] = v
        # Get the property's docstring and add it as a comment
        docstring = getattr(config.__class__, k).fget.__doc__
        commented_config_dict.yaml_set_comment_before_after_key(k, before=docstring)

    if config_path.exists() and not regenerate:
        insert = False
        with open(config_path, 'r') as file:
            saved_config_dict = yaml.load(file)
            if saved_config_dict is None:
                saved_config_dict = {}
            for k, v in config_dict.items():
                if k not in saved_config_dict:
                    saved_config_dict[k] = v
                    insert = True

        # If the config file is missing some keys, add them
        if insert:
            with open(config_path, 'w') as file:
                yaml.dump(saved_config_dict, file)
    else:
        with open(config_path, 'w') as file:
            yaml.dump(commented_config_dict, file)


def load_config_file(path=Path(os.path.expanduser('~')) / '.lynsedb_configs.yaml'):
    _config = Config()
    yaml = YAML()
    if not isinstance(path, Path):
        raise ValueError('path must be a Path object')

    if not path.exists():
        generate_config_file(regenerate=False)

    with open(path, 'r') as file:
        config_dict = yaml.load(file)
        if config_dict is None:
            config_dict = {}
        
        for key, value in config_dict.items():
            if "PATH" in key and isinstance(value, str):
                value = Path(value)
            setattr(_config, f'_{key}', value)

    return _config


config = load_config_file()
get_all_configs = config.get_all_configs


# Collection namespace, which will be used to store shared variables
# or information between different components of the collection
class CollectionNamespace:
    def __init__(self, name):
        self.name = name
        self.namespace = {}

    def get(self, name):
        return self.namespace.get(name)

    def set(self, name, value):
        self.namespace[name] = value

    def delete(self, name):
        del self.namespace[name]


class NamespaceManager:
    def __init__(self):
        self.collections_namespace = {}

    def add_namespace(self, name, namespace: CollectionNamespace):
        self.collections_namespace[name] = namespace

    def get_namespace(self, name):
        return self.collections_namespace.get(name)

    def delete_namespace(self, name):
        del self.collections_namespace[name]


collections_namespace = NamespaceManager()



"""session.py: this file is used to manage the database insertion operations."""
from typing import Union, List, Tuple

import numpy as np


class DataOpsSession:
    def __init__(self, db):
        """
        A class to manage the database insertion operations.

        Parameters:
            db: The database to be managed.
        """
        self.db = db

    def __enter__(self):
        return self.db

    def _commit(self):
        handler = getattr(self.db, '_matrix_serializer', self.db)
        if not handler.COMMIT_FLAG:
            self.db.commit()

    def __exit__(self, exc_type, exc_val, exc_tb):
        # If an exception occurred, rollback the transaction
        if exc_type:
            self._commit()
            raise exc_type(exc_val).with_traceback(exc_tb)

        self._commit()

        return False

    def add_item(self, vector: Union[np.ndarray, list], id: int, *,
                 field: dict = None, buffer_size: int = True) -> int:
        """
        Add a single vector to the collection.

        It is recommended to use incremental ids for best performance.

        Parameters:
            vector (np.ndarray): The vector to be added.
            id (int): The ID of the vector.
            field (dict, optional, keyword-only): The field of the vector. Default is None.
                If None, the field will be set to an empty string.
            buffer_size (int or bool or None): The buffer size for the storage worker. Default is True.
                
                - If None, the vector will be directly written to the disk.
                - If True, the buffer_size will be set to chunk_size,
                    and the vectors will be written to the disk when the buffer is full.
                - If False, the vector will be directly written to the disk.
                - If int, when the buffer is full, the vectors will be written to the disk.

        Returns:
            int: The ID of the added vector.

        Raises:
            ValueError: If the vector dimensions don't match or the ID already exists.
        """
        return self.db.add_item(vector, id=id, field=field, buffer_size=buffer_size)

    def bulk_add_items(
            self, vectors: Union[List[Tuple[np.ndarray, int, dict]], List[Tuple[np.ndarray, int]]],
            **kwargs
    ):
        """
        Bulk add vectors to the collection in batches.

        It is recommended to use incremental ids for best performance.

        Parameters:
            vectors (list or tuple): A list or tuple of vectors to be saved.
                Each vector can be a tuple of (vector, id, field).

        Returns:
            list: A list of indices where the vectors are stored.
        """
        return self.db.bulk_add_items(vectors, **kwargs)



import numpy as np
import cloudpickle
from spinesUtils.asserts import raise_if

from .base import BaseIndex
from ..computational_layer.engines import inner_product as ip, l2sq, cosine


class _IndexSQ(BaseIndex):
    def __init__(self):
        """
        Initialize the scalar quantizer.
        """
        super().__init__()
        self.bits = np.int8
        self.n_levels = 2 ** 7
        self.range_vals = None
        self.min_vals = None
        self.max_vals = None
        self.fitted = False

        self.data = None
        self.ids = None

    def fit(self, vectors):
        """
        Fit the model with the entire dataset to determine the global minimum and maximum range values.

        Parameters:
            vectors (np.ndarray): The input data.
        """
        self.min_vals = np.min(vectors, axis=0)
        self.max_vals = np.max(vectors, axis=0)
        self.range_vals = self.max_vals - self.min_vals
        self.fitted = True

    def encode(self, vectors):
        raise_if(ValueError, not self.fitted, 'The model must be fitted before encoding.')

        epsilon = 1e-9
        n_levels_minus_1 = self.n_levels - 1
        range_vals = self.range_vals
        min_vals = self.min_vals

        quantized = (vectors - min_vals) / (range_vals + epsilon) * n_levels_minus_1
        quantized = np.clip(quantized, -n_levels_minus_1, n_levels_minus_1).astype(self.bits)

        return quantized

    def fit_transform(self, vectors, ids):
        """
        Fit the model to the data and encode the input data to quantized vectors.

        Parameters:
            vectors (np.ndarray): The input data.
            ids (np.ndarray): The input ids.

        Returns:
            np.ndarray: The quantized vectors.
        """
        if not self.fitted:
            self.fit(vectors)

        data = self.encode(vectors)
        self.data = data if self.data is None else np.vstack((self.data, data))
        self.ids = ids if self.ids is None else np.concatenate((self.ids, ids))

    def decode(self, quantized_vectors):
        """
        Decode the quantized vectors to the original data.

        Parameters:
            quantized_vectors (np.ndarray): The quantized vectors.

        Returns:
            np.ndarray: The original data.
        """
        raise_if(ValueError, not self.fitted, 'The model must be fitted before decoding.')
        n_levels_minus_1 = self.n_levels - 1
        range_vals = self.range_vals
        min_vals = self.min_vals

        decoded = (quantized_vectors.astype(np.float32) / n_levels_minus_1) * range_vals + min_vals

        return decoded

    def _pre_select_topk(self, vec, data, ids, topk):
        """
        Pre-select the top-k indices and values.
        """
        encoded_vec = self.encode(vec)
        pre_select_ids = cosine(encoded_vec, data, n=topk, use_simd=False)[0]
        return self.decode(data[pre_select_ids]), ids[pre_select_ids]

    @staticmethod
    def _input_dtype_convert(original_vec, decoded_data):
        if original_vec.dtype != decoded_data.dtype:
            original_vec = original_vec.astype(decoded_data.dtype)
        return original_vec

    def compute_l2_distance(self, vec, mat, topk):
        """
        Compute the L2 distance between the query vector and the quantized matrix.

        Parameters:
            vec (np.ndarray): The query vector.
            mat (np.ndarray): The quantized matrix.
            topk (int): The number of nearest neighbors to return.


        Returns:
            np.ndarray: The L2 distance between the query vector and the quantized matrix.
        """
        return l2sq(vec, mat, n=topk, use_simd=True)

    def compute_inner_product(self, vec, mat, topk):
        """
        Compute the inner product between the query vector and the quantized matrix.

        Parameters:
            vec (np.ndarray): The query vector.
            mat (np.ndarray): The quantized matrix.
            topk (int): The number of nearest neighbors to return.

        Returns:
            np.ndarray: The inner product between the query vector and the quantized matrix.
        """
        return ip(self._input_dtype_convert(vec, mat), mat, n=topk, use_simd=False)

    def compute_cosine_similarity(self, vec, mat, topk):
        """
        Compute the cosine similarity between the query vector and the quantized matrix.

        Parameters:
            vec (np.ndarray): The query vector.
            mat (np.ndarray): The quantized matrix.
            topk (int): The number of nearest neighbors to return.

        Returns:
            np.ndarray: The cosine similarity between the query vector and the quantized matrix.
        """
        return cosine(self._input_dtype_convert(vec, mat), mat, n=topk, use_simd=False)

    def save(self, filepath):
        """
        Save the model to a file.

        Parameters:
            filepath (str or Pathlike): The name of the file to save the model to.
        """
        raise_if(ValueError, not self.fitted, 'The model must be fitted before saving.')
        static_vars = {
            'range_vals': self.range_vals,
            'min_vals': self.min_vals,
            'max_vals': self.max_vals,
            'fitted': self.fitted
        }
        try:
            with open(filepath, 'wb') as file:
                cloudpickle.dump(static_vars, file)
        except IOError as e:
            print(f"Error saving model: {e}")

    def save_data(self, filepath1, filepath2):
        """
        Save the data to a file.

        Parameters:
            filepath1 (str or Pathlike): The name of the file to save the data to.
            filepath2 (str or Pathlike): The name of the file to save the ids to.
        """
        if self.data is None:
            return
        try:
            with open(filepath1, 'wb') as file:
                np.save(file, self.data)
            with open(filepath2, 'wb') as file:
                np.save(file, self.ids)
        except IOError as e:
            print(f"Error saving data: {e}")

    def load(self, filepath):
        """
        Load the model from a file.

        Parameters:
            filepath (str or Pathlike): The name of the file to load the model from.

        Returns:
            _IndexSQ: The loaded model.
        """
        try:
            with open(filepath, 'rb') as file:
                static_vars = cloudpickle.load(file)

            self.fitted = static_vars['fitted']
            self.range_vals = static_vars['range_vals']
            self.min_vals = static_vars['min_vals']
            self.max_vals = static_vars['max_vals']
            return self

        except IOError as e:
            print(f"Error loading model: {e}")
            return None


class IndexSQL2sq(_IndexSQ):
    name = 'IndexSQL2sq'

    def __init__(self):
        """
        Initialize the scalar quantizer.
        """
        super().__init__()
        self._register_distance("ScalarQuantizerL2sq")

    def search(self, original_vec=None, top_k=10, rescore_multiplier=2, subset_indices=None):
        """
        Search for the nearest neighbors of the input data.
        
        Parameters:
            original_vec (np.ndarray): The original vector.
            top_k (int): The number of nearest neighbors to return.
            rescore_multiplier (int): The rescore multiplier.
            subset_indices (np.ndarray): The subset indices.
            
        Returns:
            (np.ndarray, np.ndarray): The top k indices and their distances.
        """
        encoded_vec = self.encode(original_vec)

        encoded_data = self.data
        ids = self.ids
        if subset_indices is not None:
            filtered_indices = np.isin(ids, subset_indices, assume_unique=True)
            encoded_data = encoded_data[filtered_indices]
            ids = ids[filtered_indices]

        _ids1, _ = self.compute_l2_distance(encoded_vec, encoded_data, topk=top_k * rescore_multiplier)

        decoded_data = self.decode(encoded_data[_ids1])

        _ids2, scores = self.compute_l2_distance(original_vec, decoded_data, topk=top_k)
        return ids[_ids1][_ids2], scores


class IndexSQIP(_IndexSQ):
    name = 'IndexSQIP'

    def __init__(self):
        """
        Initialize the scalar quantizer.
        """
        super().__init__()
        self._register_distance("ScalarQuantizerIP")

    def search(self, original_vec=None, top_k=10, rescore_multiplier=2, subset_indices=None):
        """
        Search for the nearest neighbors of the input data.
        """
        encoded_data = self.data
        ids = self.ids
        if subset_indices is not None:
            filtered_indices = np.isin(ids, subset_indices, assume_unique=True)
            encoded_data = encoded_data[filtered_indices]
            ids = ids[filtered_indices]

        decoded_data, ids = self._pre_select_topk(original_vec, encoded_data, ids,
                                                  top_k * rescore_multiplier)
        _ids, scores = self.compute_inner_product(original_vec, decoded_data, topk=top_k)
        return ids[_ids], scores * -1


class IndexSQCos(_IndexSQ):
    name = 'IndexSQCos'

    def __init__(self):
        """
        Initialize the scalar quantizer.
        """
        super().__init__()
        self._register_distance("ScalarQuantizerCosine")

    def search(self, original_vec=None, top_k=10, subset_indices=None):
        """
        Search for the nearest neighbors of the input data.
        """
        encoded_data = self.data
        ids = self.ids
        if subset_indices is not None:
            filtered_indices = np.isin(ids, subset_indices, assume_unique=True)
            encoded_data = encoded_data[filtered_indices]
            ids = ids[filtered_indices]

        encoded_vec = self.encode(original_vec)
        _ids, scores = self.compute_cosine_similarity(encoded_vec, encoded_data, topk=top_k)
        return ids[_ids], scores



from collections import OrderedDict

import numpy as np

from ..core_components.locks import ThreadLock


class LimitedDict:
    def __init__(self, max_size):
        if not isinstance(max_size, int):
            raise ValueError('max_size must be an integer')

        if max_size == -1:
            self.max_size = np.inf
        elif max_size < 0:
            raise ValueError('max_size must be a positive integer or -1')
        else:
            self.max_size = max_size

        self.cache = OrderedDict()
        self.lock = ThreadLock()

    def __setitem__(self, key, value):
        if self.max_size == 0:
            return

        with self.lock:
            if key in self.cache:
                del self.cache[key]
            self.cache[key] = value

            if self.cache is not None and len(self.cache) > self.max_size:
                self.cache.popitem(last=False)

    def __getitem__(self, key):
        if self.max_size == 0:
            return

        if key in self.cache:
            value = self.cache[key]
        else:
            raise KeyError('Key not found')

        return value

    @property
    def is_reached_max_size(self):
        if self.max_size == 0:
            return True

        with self.lock:
            return len(self.cache) == self.max_size

    def get(self, key, default=None):
        if self.max_size == 0:
            return default

        if key not in self.cache:
            return default
        return self.__getitem__(key)

    def clear(self):
        with self.lock:
            self.cache.clear()

    def keys(self):
        if self.max_size == 0:
            return []

        with self.lock:
            return self.cache.keys()

    def pop(self, key, default=None):
        if self.max_size == 0:
            return default

        with self.lock:
            return self.cache.pop(key, default)

    def __contains__(self, key):
        with self.lock:
            return key in self.cache

    def __len__(self):
        with self.lock:
            return len(self.cache)

    def __repr__(self):
        with self.lock:
            return repr(self.cache)



import numpy as np


class FastSort:
    def __init__(self, data):
        self.data = np.asarray(data) if not isinstance(data, np.ndarray) else data
        self.data = self.data.squeeze()

    def topk(self, k, ascending=True):
        axis = 0
        k = min(k, self.data.shape[axis])
        
        if ascending:
            partitioned_indices = np.argpartition(self.data, k-1, axis=axis)
            topk_indices = partitioned_indices[:k]
        else:
            partitioned_indices = np.argpartition(self.data, self.data.shape[1] - k, axis=axis)
            topk_indices = partitioned_indices[-k:]

        topk_values = np.take_along_axis(self.data, topk_indices, axis=axis)
        sorted_indices = np.argsort(topk_values, axis=axis)
        if not ascending:
            sorted_indices = sorted_indices[::-1]

        topk_indices = np.take_along_axis(topk_indices, sorted_indices, axis=axis)

        topk_values = np.take_along_axis(self.data, topk_indices, axis=axis)
        return topk_indices, topk_values



import cloudpickle
import numpy as np

from .base import BaseIndex
from ..computational_layer.engines import jaccard, hamming, inner_product


class _IndexBinary(BaseIndex):
    name = 'IndexBinary'

    def __init__(self):
        """
        Initialize the binary scalar quantizer.

        """
        super().__init__()

        self.means = None
        self.data = None
        self.ids = None

    @staticmethod
    def _convert_input(vectors):
        if not isinstance(vectors, np.ndarray):
            vectors = np.asarray(vectors)
        return vectors

    def fit(self, vectors):
        """
        Fit the model to the data.

        Parameters:
            vectors (np.ndarray): The input data.

        Returns:
            None
        """
        vectors = self._convert_input(vectors)
        self.means = np.mean(vectors, axis=0)

    def encode(self, vectors):
        """
        Encode the input data to binary vectors.

        Parameters:
            vectors (np.ndarray): The input data.

        Returns:
            np.ndarray: The binary vectors.
        """
        vectors = self._convert_input(vectors)

        if vectors.ndim == 1:
            vectors = vectors.reshape(1, -1)
        binary_vectors = np.where(vectors > self.means, 1, 0)
        packed_bits = np.packbits(binary_vectors, axis=1)
        return packed_bits

    def fit_transform(self, vectors, ids):
        """
        Fit the model to the data and encode the input data to binary vectors.

        Parameters:
            vectors (np.ndarray): The input data.
            ids (np.ndarray): The input ids.

        Returns:
            np.ndarray: The binary vectors.
        """
        if self.means is None:
            self.fit(vectors)

        data = self.encode(vectors)
        self.data = data if self.data is None else np.vstack([self.data, data])
        self.ids = ids if self.ids is None else np.hstack([self.ids, ids])

    def _generate_encoded_data(self, encoded_data):
        if encoded_data is None:
            encoded_data = self.data
        return encoded_data

    def rescore(self, original_vec, top_k, decoded_sq_data, ids):
        original_vec = self._convert_input(original_vec)

        if original_vec.dtype != decoded_sq_data.dtype:
            query = original_vec.astype(decoded_sq_data.dtype)
        else:
            query = original_vec

        ip_ids, ip_score = inner_product(query, decoded_sq_data, n=top_k, use_simd=True)

        return ids[ip_ids], ip_score

    def _search(self, original_vec, original_data, encoded_vec, encoded_data, top_k,
                rescore, rescore_multiplier, subset_indices, distance):
        original_vec = self._convert_input(original_vec)
        encoded_data = self._generate_encoded_data(encoded_data)

        ids = self.ids

        if subset_indices is not None:
            filtered_indices = np.isin(self.ids, subset_indices, assume_unique=True)
            encoded_data = encoded_data[filtered_indices]
            ids = ids[filtered_indices]

        encoded_vec, encoded_data = super().check_and_encode(original_vec=original_vec, encoded_vec=encoded_vec,
                                                             original_data=original_data, encoded_data=encoded_data)

        if distance == 'Jaccard':
            distance_func = jaccard
        elif distance == 'Hamming':
            distance_func = hamming
        else:
            raise ValueError(f"Invalid distance: {distance}")

        sort_topk = top_k * rescore_multiplier if rescore else top_k
        _ids, scores = distance_func(encoded_vec, encoded_data, n=sort_topk, use_simd=True)

        if rescore:
            if subset_indices is not None:
                sq8_data = self.sq8_data[_ids]
            else:
                sq8_data = self.sq8_data

            decoded_sq_data = self.sq8_decode(sq8_data[_ids])
            return self.rescore(original_vec, top_k, decoded_sq_data, ids[_ids])

        return ids[_ids], scores if distance == 'Jaccard' else scores / original_vec.squeeze().shape[0]

    def save(self, filepath):
        """
        Save the model to a file.

        Parameters:
            filepath (str or Pathlike): The name of the file to save the model to.
        """
        static_vars = {
            'means': self.means
        }
        with open(filepath, 'wb') as f:
            cloudpickle.dump(static_vars, f)

    def save_data(self, filepath1, filepath2):
        """
        Save the data to a file.

        Parameters:
            filepath1 (str or Pathlike): The name of the file to save the data to.
            filepath2 (str or Pathlike): The name of the file to save the ids to.
        """
        if self.data is None:
            return
        with open(filepath1, 'wb') as f:
            np.save(f, self.data)
        with open(filepath2, 'wb') as f:
            np.save(f, self.ids)

    def load(self, filepath):
        """
        Load the model from a file.

        Parameters:
            filepath (str or Pathlike): The name of the file to load the model from.

        Returns:
            The loaded model.
        """
        try:
            with open(filepath, 'rb') as file:
                means = cloudpickle.load(file)
            self.means = means['means']

            return self
        except IOError as e:
            print(f"Error loading model: {e}")
            return None


class IndexBinaryJaccard(_IndexBinary):
    name = 'IndexBinaryJaccard'

    def __init__(self):
        """
        Initialize the binary scalar quantizer.

        """
        super().__init__()
        self._register_distance("Jaccard")

    def search(self, original_vec=None, encoded_vec=None, original_data=None,
               encoded_data=None, top_k=10, rescore=False, rescore_multiplier=2,
               subset_indices=None):
        """
        Search for the nearest neighbors of the input data.

        Parameters:
            original_vec (np.ndarray): The input data.
            encoded_vec (np.ndarray): The encoded data.
            original_data (np.ndarray): The input data.
            encoded_data (np.ndarray): The encoded data.
            top_k (int): The number of top elements to return.
            rescore (bool): Whether to rescore the results. If True, the distance will be set to InnerProduct.
            rescore_multiplier (int): The multiplier for the rescore.
            subset_indices (np.ndarray): The indices of the subset to search.

        Returns:
            (np.ndarray, np.ndarray): The top k values and their indices.
        """
        return self._search(original_vec, original_data, encoded_vec, encoded_data, top_k, rescore,
                            rescore_multiplier, subset_indices, 'Jaccard')


class IndexBinaryHamming(_IndexBinary):
    name = 'IndexBinaryHamming'

    def __init__(self):
        """
        Initialize the binary scalar quantizer.

        """
        super().__init__()
        self._register_distance("Hamming")

    def search(self, original_vec=None, encoded_vec=None, original_data=None,
               encoded_data=None, top_k=10, rescore=False, rescore_multiplier=2, subset_indices=None):
        """
        Search for the nearest neighbors of the input data.

        Parameters:
            original_vec (np.ndarray): The input data.
            encoded_vec (np.ndarray): The encoded data.
            original_data (np.ndarray): The input data.
            encoded_data (np.ndarray): The encoded data.
            top_k (int): The number of top elements to return.
            rescore (bool): Whether to rescore the results. If True, the distance will be set to InnerProduct.
            rescore_multiplier (int): The multiplier for the rescore.
            subset_indices (np.ndarray): The indices of the subset to search.

        Returns:
            (np.ndarray, np.ndarray): The top k values and their indices.
        """
        return self._search(original_vec, original_data, encoded_vec, encoded_data, top_k, rescore,
                            rescore_multiplier, subset_indices, 'Hamming')



from pathlib import Path

from setuptools import find_packages, setup


def read_requirements(path):
    return list(Path(path).read_text().splitlines())


reqs = read_requirements(Path('.').parent.joinpath("requirements.txt"))

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='LynseDB',
    version="0.1.1",
    description='A pure Python-implemented, lightweight, server-optional, '
                'multi-end compatible, vector database deployable locally or remotely.',
    keywords='vector database',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    python_requires=">=3.9",
    url='https://github.com/BirchKwok/MinVectorDB',
    author='Birch Kwok',
    author_email='birchkwok@gmail.com',
    install_requires=reqs,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'lynse=lynse.api.http_api.http_api.app:main',
        ],
    },
)



import json
import uuid
from threading import Thread

import msgpack

from flask import request, jsonify, Response, Blueprint

from ....configs.config import config
from ....core_components.fields_cache import IndexSchema
from ....core_components.limited_dict import LimitedDict
from ....core_components.safe_dict import SafeDict

collection_ops = Blueprint('collection_ops', __name__)

data_dict = LimitedDict(max_size=1000)
tasks = SafeDict()

root_path = config.LYNSE_DEFAULT_ROOT_PATH


@collection_ops.before_request
def handle_msgpack_input():
    if request.headers.get('Content-Type') == 'application/msgpack':
        data = msgpack.unpackb(request.get_data(), raw=False)
        request.decoded_data = data


@collection_ops.route('/', methods=['GET'])
def index():
    return Response(json.dumps({'status': 'success', 'message': 'LynseDB HTTP API'}), mimetype='application/json')


@collection_ops.route('/required_collection', methods=['POST'])
def required_collection():
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    if 'chunk_size' not in data:
        data['chunk_size'] = 100000
    if 'dtypes' not in data:
        data['dtypes'] = 'float32'
    if 'use_cache' not in data:
        data['use_cache'] = True
    if 'n_threads' not in data:
        data['n_threads'] = 10
    if 'warm_up' not in data:
        data['warm_up'] = True
    if 'drop_if_exists' not in data:
        data['drop_if_exists'] = False
    if 'description' not in data:
        data['description'] = None
    if 'cache_chunks' not in data:
        data['cache_chunks'] = 20

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        my_vec_db.require_collection(
            collection=data['collection_name'],
            dim=data['dim'],
            chunk_size=data['chunk_size'],
            dtypes=data['dtypes'],
            use_cache=data['use_cache'],
            n_threads=data['n_threads'],
            warm_up=data['warm_up'],
            drop_if_exists=data['drop_if_exists'],
            description=data['description'],
            cache_chunks=data['cache_chunks']
        )
        return jsonify({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name']
        }}), 200
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/drop_collection', methods=['POST'])
def drop_collection():
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        my_vec_db.drop_collection(data['collection_name'])
        return jsonify({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name']}
        }), 200
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/show_collections', methods=['POST'])
def show_collections():
    """Show all collections in the database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    if 'database_name' not in data:
        return jsonify({'error': 'Missing required parameter: database_name'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collections = my_vec_db.show_collections()

        return Response(json.dumps({
            'status': 'success', 'params': {
                'database_name': data['database_name'],
                'collections': collections
            }
        }, sort_keys=False), mimetype='application/json')

    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/add_item', methods=['POST'])
def add_item():
    from ....api.native_api.high_level import LocalClient

    if request.headers.get('Content-Type') == 'application/msgpack':
        data = request.decoded_data
    else:
        data = request.json

    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        id = collection.add_item(vector=data['item']['vector'], id=data['item']['id'],
                                 field=data['item'].get('field', {}))

        return Response(json.dumps(
            {
                'status': 'success', 'params': {
                'database_name': data['database_name'],
                'collection_name': data['collection_name'], 'item': {
                    'id': data['item']['id']
                }
            }
            }, sort_keys=False),
            mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/bulk_add_items', methods=['POST'])
def bulk_add_items():
    from ....api.native_api.high_level import LocalClient

    if request.headers.get('Content-Type') == 'application/msgpack':
        data = request.decoded_data
    else:
        data = request.json

    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        items = []
        for item in data['items']:
            items.append((item['vector'], item['id'], item.get('field', None)))
        ids = collection.bulk_add_items(items)

        return Response(json.dumps({
            'status': 'success', 'params': {
                'database_name': data['database_name'],
                'collection_name': data['collection_name'], 'ids': [i['id'] for i in data['items']]
            }
        }, sort_keys=False),
            mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/search', methods=['POST'])
def search():
    """Search the database for the vectors most similar to the given vector.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient
    from ....core_components.fields_cache.filter import Filter

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    if 'k' not in data:
        data['k'] = 10

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        if data['search_filter'] is None:
            search_filter = None
        else:
            search_filter = Filter().load_dict(data['search_filter'])

        indexer = getattr(collection, '_indexer', None)
        if indexer is not None:
            index_mode_judge = 'Binary' not in indexer.index_mode
        else:
            index_mode_judge = False

        ids, scores, field = collection.search(
            vector=data['vector'], k=data['k'],
            search_filter=search_filter,
            return_fields=data.get('return_fields', False),
            rescore=data.get('rescore', False),
            rescore_multiplier=(data.get('rescore_multiplier', 2)
                                if index_mode_judge else data.get('rescore_multiplier', 10))
        )

        if ids is not None:
            ids = ids.tolist()
            scores = scores.tolist()

        return Response(json.dumps(
            {
                'status': 'success', 'params': {
                'database_name': data['database_name'],
                'collection_name': data['collection_name'], 'items': {
                    'k': data['k'], 'ids': ids, 'scores': scores,
                    'fields': field,
                }
            }
            }, sort_keys=False),
            mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


def commit_task(data):
    from ....api.native_api.high_level import LocalClient

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        if 'items' in data:
            items = []
            for item in data['items']:
                items.append((item['vector'], item['id'], item.get('field', {})))

            collection.bulk_add_items(items)

        collection.commit()

        return {'status': 'Success', 'result': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name']
        }}
    except KeyError as e:
        return {'status': 'Error', 'message': f' missing required parameter {e}'}
    except Exception as e:
        return {'status': 'Error', 'message': str(e)}


@collection_ops.route('/commit', methods=['POST'])
def commit():
    if request.headers.get('Content-Type') == 'application/msgpack':
        data = request.decoded_data
    else:
        data = request.json

    if not data:
        return jsonify({'error': 'No data provided'}), 400

    task_id = str(uuid.uuid4())
    tasks[task_id] = {'status': 'Processing'}

    def execute_task():
        result = commit_task(data)
        tasks[task_id] = result

    thread = Thread(target=execute_task)
    thread.start()

    return jsonify({'task_id': task_id}), 202


@collection_ops.route('/status/<task_id>', methods=['GET'])
def get_status(task_id):
    task = tasks.get(task_id)
    if not task:
        return jsonify({'error': 'Task not found'}), 404

    return jsonify(task)


@collection_ops.route('/collection_shape', methods=['POST'])
def collection_shape():
    """Get the shape of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'shape': collection.shape
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/get_collection_status_report', methods=['POST'])
def get_collection_status_report():
    """Get the status report of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        db_report = {'DATABASE STATUS REPORT': {
            'Database shape': (
                0, collection._matrix_serializer.dim) if collection._matrix_serializer.IS_DELETED else collection.shape,
            'Database last_commit_time': collection._matrix_serializer.last_commit_time,
            'Database commit status': collection._matrix_serializer.COMMIT_FLAG,
            'Database use_cache': collection._use_cache,
            'Database status': 'DELETED' if collection._matrix_serializer.IS_DELETED else 'ACTIVE'
        }}

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'status_report': db_report
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/is_collection_exists', methods=['POST'])
def is_collection_exists():
    """Check if a collection exists.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'exists': data['collection_name'] in my_vec_db.show_collections()
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/get_collection_config', methods=['POST'])
def get_collection_config():
    """Get the configuration of a collection.

    Returns:
        dict: The status of the operation.
    """
    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        config_json_path = root_path / data['database_name'] / 'collections.json'
        with open(config_json_path, 'r') as file:
            collections = json.load(file)
            collection_config = collections[data['collection_name']]

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'config': collection_config
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/update_commit_msg', methods=['POST'])
def update_commit_msg():
    """Save the commit message of a collection.

    Returns:
        dict: The status of the operation.
    """
    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        if (root_path / data['database_name'] / 'commit_msg.json').exists():
            with open(root_path / data['database_name'] / 'commit_msg.json', 'r') as file:
                commit_msg = json.load(file)
                commit_msg[data['collection_name']] = data
        else:
            commit_msg = {data['collection_name']: data}

        with open(root_path / data['database_name'] / 'commit_msg.json', 'w') as file:
            json.dump(commit_msg, file)

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name']
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/get_commit_msg', methods=['POST'])
def get_commit_msg():
    """Get the commit message of a collection.

    Returns:
        dict: The status of the operation.
    """
    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        if (root_path / data['database_name'] / 'commit_msg.json').exists():
            with open(root_path / data['database_name'] / 'commit_msg.json', 'r') as file:
                commit_msg = json.load(file)
                commit_msg = commit_msg.get(data['collection_name'], None)
        else:
            commit_msg = 'No commit message found for this collection'

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'commit_msg': commit_msg
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/update_collection_description', methods=['POST'])
def update_collection_description():
    """Update the description of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        my_vec_db.update_collection_description(data['collection_name'], data['description'])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'description': data['description']
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/update_description', methods=['POST'])
def update_description():
    """Update the description of the database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        collection.update_description(data['description'])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'description': data['description']
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/show_collections_details', methods=['POST'])
def show_collections_details():
    """Show all collections in the database with details.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collections_details = my_vec_db.show_collections_details()
        return Response(json.dumps({
            'status': 'success', 'params': {
                'database_name': data['database_name'],
                'collections': collections_details.to_dict()
            }
        },
            sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/build_index', methods=['POST'])
def build_index():
    """Build the index of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        collection.build_index(index_mode=data.get('index_mode', 'IVF-FLAT'))

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'index_mode': data.get('index_mode', 'IVF-FLAT')
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/remove_index', methods=['POST'])
def remove_index():
    """Remove the index of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        collection.remove_index()

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name']
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/head', methods=['POST'])
def head():
    """Get the first n items of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    if 'n' not in data:
        data['n'] = 5

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        head = collection.head(n=data['n'])
        head = (head[0].tolist(), head[1].tolist(), head[2])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'head': head
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/tail', methods=['POST'])
def tail():
    """Get the last n items of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    if 'n' not in data:
        data['n'] = 5

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        tail = collection.tail(n=data['n'])
        tail = (tail[0].tolist(), tail[1].tolist(), tail[2])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'tail': tail
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/read_by_only_id', methods=['POST'])
def read_by_only_id():
    """Read the item by only id.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        vector, id, field = collection.read_by_only_id(data['id'])
        vector = vector.tolist()
        id = id.tolist()

        item = (vector, id, field)
        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'], 'item': item
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/get_collection_path', methods=['POST'])
def get_collection_path():
    """Get the path of a database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'collection_path': collection._database_path}}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/query', methods=['POST'])
def query():
    """Query the database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        result = collection.query(data['query_filter'], data['filter_ids'], data['return_ids_only'])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'result': result}}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500

@collection_ops.route('/query_vectors', methods=['POST'])
def query_vectors():
    """Query the database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient   

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        result = collection.query_vectors(data['query_filter'], data['filter_ids'])
        result = (result[0].tolist(), result[1].tolist(), result[2])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'result': result}}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/build_field_index', methods=['POST'])
def build_field_index():
    """Build the index of a field.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        schema = IndexSchema().load_from_dict(data['schema'])

        collection.build_field_index(schema=schema,
                                     rebuild_if_exists=data['rebuild_if_exists'])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name']
        }}, sort_keys=False), mimetype='application/json')
    except KeyError as e:
        return jsonify({'error': f'Missing required parameter {e}'}), 400
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/list_fields', methods=['POST'])
def list_fields():
    """List all fields of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])
        fields = collection.list_fields()

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'fields': fields
        }}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500
    

@collection_ops.route('/list_field_index', methods=['POST'])
def list_field_index():
    """List all field indexes of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        field_indices = collection.list_field_index()

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'field_indices': field_indices}}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/remove_field_index', methods=['POST'])
def remove_field_index():
    """Remove the index of a field.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        collection.remove_field_index(data['field_name'])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'field_name': data['field_name']
        }}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/remove_all_field_indices', methods=['POST'])
def remove_all_field_indices():
    """Remove all field indices of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        collection.remove_all_field_indices()

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name']
        }}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/index_mode', methods=['POST'])
def index_mode():
    """Get the index mode of a collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        index_mode = collection.index_mode

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'index_mode': index_mode
        }}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@collection_ops.route('/is_id_exists', methods=['POST'])
def is_id_exists():
    """Check if an ID exists in the database.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        is_id_exists = collection.is_id_exists(data['id'])

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'id': data['id'],
            'is_id_exists': is_id_exists
        }}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500
    

@collection_ops.route('/max_id', methods=['POST'])
def max_id():
    """Get the maximum ID in the collection.

    Returns:
        dict: The status of the operation.
    """
    from ....api.native_api.high_level import LocalClient

    data = request.json
    if not data:
        return jsonify({'error': 'No data provided'}), 400

    try:
        my_vec_db = LocalClient(root_path=root_path / data['database_name'])
        collection = my_vec_db.get_collection(data['collection_name'])

        max_id = int(collection.max_id)

        return Response(json.dumps({'status': 'success', 'params': {
            'database_name': data['database_name'],
            'collection_name': data['collection_name'],
            'max_id': max_id
        }}, sort_keys=False), mimetype='application/json')
    except Exception as e:
        return jsonify({'error': str(e)}), 500
    


import json
import shutil
import uuid
from functools import partial
from pathlib import Path
from typing import Union

from spinesUtils.asserts import ParameterTypeAssert

from ...api import logger
from ...utils.utils import unavailable_if_deleted


class _Register:
    """
    A class for registering the collections to local database path.
    """

    def __init__(self, root_path: str) -> None:
        self.root_path = Path(root_path)

    def register_collection(self, collection: str, **kwargs):
        """
        Save the collection name to the local database path.

        Parameters:
            collection (str): The name of the collection.
        """
        if not (self.root_path / 'collections.json').exists():
            with open(self.root_path / 'collections.json', 'w') as f:
                json.dump({collection: kwargs}, f)
        else:
            collections = self.get_collections_details()

            if collection not in collections:
                collections[collection] = kwargs

                with open(self.root_path / 'collections.json', 'w') as f:
                    json.dump(collections, f)

    def deregister_collection(self, collection: str):
        """
        Delete the collection name from the local database path.

        Parameters:
            collection (str): The name of the collection.
        """
        if not (self.root_path / 'collections.json').exists():
            return

        collections = self.get_collections_details()

        if collection in collections:
            del collections[collection]

            with open(self.root_path / 'collections.json', 'w') as f:
                json.dump(collections, f)

    def get_collections_details(self) -> dict:
        """
        Show the collections in the database.

        Returns:
            list: The list of collections in the database.
        """
        if not (self.root_path / 'collections.json').exists():
            return {}

        with open(self.root_path / 'collections.json', 'r') as f:
            collections = json.load(f)

        for collection in list(collections.keys()):
            if not (self.root_path / collection).exists():
                del collections[collection]
                with open(self.root_path / 'collections.json', 'w') as f:
                    json.dump(collections, f)

        return collections

    def show_collections(self) -> list:
        """
        Show the collections in the database.

        Returns:
            list: The list of collections in the database.
        """
        return list(self.get_collections_details().keys())

    def show_collections_details(self):
        """
        Show the collections in the database.

        Returns:
            list: The list of collections in the database.
        """
        details = self.get_collections_details()

        # hide the database path
        for collection in details:
            if "database_path" in details[collection]:
                del details[collection]["database_path"]

        try:
            import pandas as pd
            details = pd.DataFrame(details).T
        except ImportError:
            ...

        return details

    @ParameterTypeAssert({
        'collection': str,
        'description': (None, str, int, float, bool)
    })
    def update_description(self, collection: str, description: Union[None, str, int, float, bool]):
        """
        Update the description of the collection.

        Parameters:
            collection (str): The name of the collection.
            description (None or str or int or float or bool): The description of the collection.
        """
        collections = self.get_collections_details()

        if collection in collections:
            collections[collection]['description'] = description

            with open(self.root_path / 'collections.json', 'w') as f:
                json.dump(collections, f)

    def __contains__(self, item):
        return item in self.get_collections_details()


class LocalClient:
    """
    A singleton class for the local LynseDB client.
    Using the LocalClient class, users can create and access database instances, 
    as well as operate on the collections within them (Low-Level API).
    This class is thread-safe only. Using it in multiple processes will result in a race condition.
    """
    _instance = None
    _last_root_path = None

    def __new__(cls, root_path: Union[Path, str]):
        """
        Create a new instance or return the existing instance of the class.

        Parameters:
            root_path (Path or str): The root path of the database.

        Returns:
            LocalClient (LocalClient): The instance of the class.
        """
        if cls._instance is not None and cls._last_root_path != root_path:
            cls._instance = None
            cls._last_root_path = root_path

        if cls._instance is None:
            cls._instance = super(LocalClient, cls).__new__(cls)
            cls._instance._init(root_path)

            cls._last_root_path = root_path

        Path(cls._last_root_path).mkdir(parents=True, exist_ok=True)

        if not (Path(cls._last_root_path) / '.fingerprint').exists():
            # create a database fingerprint
            with open(Path(cls._last_root_path) / '.fingerprint', 'w') as f:
                f.write(uuid.uuid4().hex)

        return cls._instance

    def _init(self, root_path: Union[Path, str]):
        """
        Initialize the vector database.
        """
        self._root_path = Path(root_path).absolute()

        self._register = _Register(root_path)
        self._collections = {}
        self.STATUS = 'INITIALIZED'

    def require_collection(
            self,
            collection: str,
            dim: int = None,
            chunk_size: int = 100_000,
            dtypes: str = 'float32',
            use_cache: bool = True,
            n_threads: Union[int, None] = 10,
            warm_up: bool = False,
            drop_if_exists: bool = False,
            description: str = None,
            cache_chunks: int = 20
    ):
        """Create or load a collection in the database.

        Parameters:
            collection (str): The name of the collection.
            dim (int): Dimension of the vectors. Default is None.
                When creating a new collection, the dimension of the vectors must be specified.
                When loading an existing collection, the dimension of the vectors is automatically loaded.
            chunk_size (int): The size of each data chunk. Default is 100_000.
            dtypes (str): The data type of the vectors. Default is 'float32'.
                Options are 'float16', 'float32' or 'float64'.
            use_cache (bool): Whether to use cache for search. Default is True.
            n_threads (int): The number of threads to use for parallel processing. Default is 10.
            warm_up (bool): Whether to warm up the database. Default is False.
            drop_if_exists (bool): Whether to drop the collection if it already exists. Default is False.
            description (str): A description of the collection. Default is None.
                The description is limited to 500 characters.
            cache_chunks (int): The number of chunks to cache. Default is 20.

        Raises:
            ValueError: If `chunk_size` is less than or equal to 1.
        """
        from ...api.native_api.low_level import ExclusiveDB

        collection_path = self._root_path / collection

        if description is not None and not isinstance(description, str):
            raise ValueError('Description must be a string')
        elif description is not None and len(description) > 500:
            raise ValueError('Description must be less than 500 characters')

        if collection in self._register:
            if drop_if_exists:
                self.drop_collection(collection)
                logger.info(f"Collection '{collection}' already exists. Dropped.")
            else:
                collection_details = self._register.get_collections_details()[collection]
                dim = collection_details['dim']
                logger.info(f"Collection '{collection}' already exists. Loaded.")

        if chunk_size <= 1:
            raise ValueError('chunk_size must be greater than 1')

        self._collections[collection] = ExclusiveDB(
            dim=dim, database_path=collection_path.as_posix(), chunk_size=chunk_size, dtypes=dtypes,
            use_cache=use_cache, n_threads=n_threads,
            warm_up=warm_up, cache_chunks=cache_chunks
        )

        self._register.register_collection(
            collection, dim=dim, database_path=collection_path.as_posix(), chunk_size=chunk_size, dtypes=dtypes,
            use_cache=use_cache, n_threads=n_threads,
            warm_up=warm_up, description=description, cache_chunks=cache_chunks
        )

        self._collections[collection].update_description = partial(self.update_collection_description, collection)

        return self._collections[collection]

    @unavailable_if_deleted
    def get_collection(self, collection: str, cache_chunks: int = 20, warm_up: bool = True):
        """
        Get a collection from the database.

        Parameters:
            collection (str): The name of the collection to get.
            cache_chunks (int): The number of chunks to cache. Default is 20.
            warm_up (bool): Whether to warm up the database. Default is True.

        Returns:
            (ExclusiveDB): The collection.
        """
        from ...api.native_api.low_level import ExclusiveDB

        if collection not in self._collections:
            if collection not in self._register:
                raise ValueError(f"Collection '{collection}' does not exist.")

            params = self._register.get_collections_details()[collection]
            params.update({'cache_chunks': cache_chunks, 'warm_up': warm_up})
            if 'description' in params:
                del params['description']
            self._collections[collection] = ExclusiveDB(**params)

        return self._collections[collection]

    @unavailable_if_deleted
    def copy_collection(self, deep=False):
        """
        Copy the collection.

        Parameters:
            deep (bool): Whether to make a deep copy. Default is False.
                Which means share the data with the original collection.

        Returns:
            (ExclusiveDB): The copied collection.

        Raises:
            NotImplementedError: This method is not implemented yet.
        """
        raise NotImplementedError("This method is not implemented yet.")

    @unavailable_if_deleted
    def show_collections(self):
        """
        Show the collections in the database.

        Returns:
            List: The list of collections in the database.
        """
        return self._register.show_collections()

    @unavailable_if_deleted
    def drop_collection(self, collection: str):
        """
        Delete a collection from the database.

        Parameters:
            collection (str): The name of the collection to delete.

        Returns:
            None
        """

        if collection in self._collections:
            self._collections[collection].delete()
            del self._collections[collection]
            self._register.deregister_collection(collection)
        else:
            try:
                _temp_collection = self.get_collection(collection)
                self.drop_collection(collection)
            except ValueError:
                pass

        if (self._root_path / "collections.json").exists():
            with open(self._root_path / "collections.json", "r") as f:
                collections = json.load(f)
                if collection in collections:
                    del collections[collection]

            with open(self._root_path / "collections.json", "w") as f:
                json.dump(collections, f)

    def drop_database(self):
        """
        Delete the database.

        Returns:
            None
        """
        if self.STATUS == 'DELETED':
            return

        if self._root_path.exists():
            shutil.rmtree(self._root_path)

        LocalClient._instance = None
        self.STATUS = 'DELETED'

    def update_collection_description(self, collection: str, description: Union[None, str, int, float, bool]):
        """
        Update the description of the collection.

        Parameters:
            collection (str): The name of the collection.
            description (None or str or int or float or bool): The description of the collection.

        Returns:
            None
        """
        self._register.update_description(collection, description)

    def show_collections_details(self):
        """
        Show the collections in the database.

        Returns:
            (Dict or DataFrame): The details of the collections in the database.
        """
        return self._register.show_collections_details()

    def database_exists(self):
        """
        Check if the database exists.

        Returns:
            (Bool): Whether the database exists.
        """
        return self._root_path.exists()

    @property
    def root_path(self):
        """
        Get the root path of the database.

        Returns:
            (str): The root path of the database.
        """
        return self._root_path.as_posix()

    def __repr__(self):
        return (f'{self.__class__.__name__.replace("Client", "DatabaseInstance")}(name={self._root_path.name},'
                f' exists={self.database_exists()})')

    def __str__(self):
        return self.__repr__()
